# -*- coding: utf-8 -*-

import base64
import json
from xml.etree import ElementTree as eTree
import sysactions
from fiscalpersistence import Tender
from msgbus import MBEasyContext
from typing import Dict, List, Optional

from _AposA8PrintLine import AposA8PrintLine
from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType
from _AposA8Constants import AposA8Constants

from printer import FiscalPrinter
from printer_wrapper import insert_space
from old_helper import F, get_sale_line_priced_items, round_half_away_from_zero
from clichehelper import ClicheHelper


class SatAposA8FiscalPrinter(FiscalPrinter):
    def __init__(self, mb_context, print_customer_name, cliche_helper, customer_text_field, tip_config):
        # type: (MBEasyContext, bool, ClicheHelper, str, Dict[str]) -> None
        super(SatAposA8FiscalPrinter, self).__init__(mb_context, print_customer_name, cliche_helper)
        self.customer_text_field = customer_text_field
        self.tip_config = tip_config

    def print_fiscal(self, posid, order, tenders, fiscal_request, fiscal_response, total_taxes, customer_name, url_sefaz=""):
        # type: (str, eTree.Element, list, eTree.Element, eTree.Element, dict, str, str) -> str
        xml = base64.b64decode(fiscal_request.text.strip())
        parsed_xml = eTree.XML(xml)
        n_c_fe = parsed_xml.find("infCFe/ide/nCFe").text
        model = sysactions.get_model(posid)

        report = self._fiscal_report(order, tenders, fiscal_response, n_c_fe, customer_name, model)

        qr_code = fiscal_response.find("QRCode").text.strip()
        sat_key = fiscal_response.find("satkey").text.strip()
        report.append(AposA8PrintLine(AposA8Align.CENTER, sat_key, AposA8CommandType.PRINT_BAR_CODE))
        report.append(AposA8PrintLine(AposA8Align.CENTER, '', AposA8CommandType.PRINT_BAR_CODE))
        report.append(AposA8PrintLine(AposA8Align.CENTER, qr_code, AposA8CommandType.PRINT_QR_CODE))

        text = "Consulte o QRCode pelo Aplicativo"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))
        text = "De olho na nota disponivel"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))
        text = "na AppStore (Apple) e Play Store (Android)"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        tax_text = "%s %7.2f" % ("Valor aproximado dos tributos deste cupom", (
                    float(total_taxes['nacionalfederal']['value']) + float(total_taxes['estadual']['value'])))
        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      tax_text,
                                      AposA8CommandType.PRINT_TEXT))

        federal_tax = "Fed = R$%7.2f (%.2f%%), Est = R$%7.2f (%.2f%%)" % (
            total_taxes['nacionalfederal']['value'], total_taxes['nacionalfederal']['percent'],
            total_taxes['estadual']['value'], total_taxes['estadual']['percent'])

        report.append(AposA8PrintLine(AposA8Align.CENTER, federal_tax,  AposA8CommandType.PRINT_TEXT))

        text = "(Conforme Lei Fed. 12.741/2012)"
        report.append(AposA8PrintLine(AposA8Align.CENTER, text, AposA8CommandType.PRINT_TEXT))

        return json.dumps(report, default=lambda x: x.__dict__)

    def print_cancel(self, posid, data):
        return

    def _cancel_report(self, sat_fiscal_response, nCFe, total):
        return

    def _fiscal_report(self, order, tenders, sat_fiscal_response, nCFe, customer_name, model=None):
        # type: (eTree, List[Tender], eTree, dict, str, str, Optional[eTree]) -> List[AposA8PrintLine]

        whopper_wifi_code = order.find("CustomOrderProperties/OrderProperty[@key='WHOPPER_WIFI_CODE']").get(
            "value") if order.find(
            "CustomOrderProperties/OrderProperty[@key='WHOPPER_WIFI_CODE']") is not None else None
        if whopper_wifi_code is not None:
            whopper_wifi_code = whopper_wifi_code[0:4] + " " + whopper_wifi_code[4:8]

        report = []

        for line in self.cliche_helper.get_cliche():
            report.append(AposA8PrintLine(AposA8Align.CENTER, line, AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, "EXTRATO No. %s" % nCFe, AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, "CUPOM FISCAL ELETRONICO - SAT", AposA8CommandType.PRINT_TEXT))
        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        product_header = "COD     DESCR          QTD UN    VL UNIT   TOTAL"
        report.append(AposA8PrintLine(AposA8Align.CENTER, product_header, AposA8CommandType.PRINT_TEXT))

        order_voided = order.get("state") == "VOIDED"
        items_amount = 0

        sale_lines = order.findall("SaleLine")
        sale_lines = get_sale_line_priced_items(sale_lines)

        for line in sale_lines:
            line_params = map(line.get, ("itemPrice", "multipliedQty", "productName", "partCode", "level"))
            item_price, multiplied_qty, name, part_code, level = line_params
            measure_unit = line.get("measureUnit") if "measureUnit" in line.attrib else "un"
            items_amount += 1
            quantity = F(multiplied_qty)
            item_value = round_half_away_from_zero(F(item_price) / quantity, 2)
            line = "%-7.7s %-18.18s%6.3f%2.2s%9.02f" % (part_code, name, quantity, measure_unit, item_value)
            string = ("%-38.38s %9.02f" % (line, F(item_price)))

            report.append(AposA8PrintLine(AposA8Align.CENTER, string, AposA8CommandType.PRINT_TEXT))

        if order.get("tip", 0) and float(order.get("tip")) > 0:
            order_tip = F(order.get("tip"))
            tip_line = "%-7.7s %-14.14s%5.3f%2.2s%9.02f" % (self.tip_config['TipCode'],
                                                            sysactions.translate_message(model, "TIP_GIVEN").upper(),
                                                            F(1),
                                                            "un",
                                                            order_tip)
            tip_text = ("%-38.38s %9.02f" % (tip_line, order_tip))
            report.append(AposA8PrintLine(AposA8Align.CENTER, tip_text, AposA8CommandType.PRINT_TEXT))

        factor = -1 if order_voided else 1

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        qty_total = "%-36s         %3d" % ("QTD. TOTAL DE ITENS", items_amount)
        report.append(AposA8PrintLine(AposA8Align.CENTER, qty_total, AposA8CommandType.PRINT_TEXT))

        total_amount = "%-36s   R$%7.2f" % ("VALOR TOTAL", float(order.get("totalAmount")) + float(order.get("taxTotal")))
        report.append(AposA8PrintLine(AposA8Align.CENTER, total_amount, AposA8CommandType.PRINT_TEXT))

        discount = "%-36s   R$%7.2f" % ("DESCONTO", float(order.get("discountAmount")))
        report.append(AposA8PrintLine(AposA8Align.CENTER, discount, AposA8CommandType.PRINT_TEXT))

        total_payable = "%-36s   R$%7.2f" % ("VALOR A PAGAR",
                                             float(order.get("totalAmount"))
                                             + float(order.get("taxTotal"))
                                             - float(order.get("discountAmount")))
        report.append(AposA8PrintLine(AposA8Align.CENTER, total_payable, AposA8CommandType.PRINT_TEXT))

        payment = "%-36s  %s" % ("FORMA DE PAGAMENTO", "Valor Pago")
        report.append(AposA8PrintLine(AposA8Align.CENTER, payment, AposA8CommandType.PRINT_TEXT))

        for tender in tenders:
            tender_descr = self.get_payment_description(self.mb_context, tender)
            tender_amount = tender.amount
            report.append(AposA8PrintLine(AposA8Align.CENTER, "%-36s   R$%7.2f" % (tender_descr, tender_amount),
                                          AposA8CommandType.PRINT_TEXT))

        # Get "Change" (sixth value of tender tuple) from last tender in tenders
        change = tenders[len(tenders) - 1].change

        if change > 0:
            report.append(AposA8PrintLine(AposA8Align.CENTER, "%-36s   R$%7.2f" % ("Troco", change * factor),
                                          AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))
        if self.print_customer_name:
            char_amount = 21

            content = (customer_name or '')[:(char_amount - 2)] or order.get('orderId')[-3:]
            report.append(AposA8PrintLine(AposA8Align.CENTER, "PAINEL DE RETIRADA - VOCE SERA CHAMADO POR:",
                                          AposA8CommandType.PRINT_TEXT))

            report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_HASH,
                                          AposA8CommandType.PRINT_TEXT))
            report.append(AposA8PrintLine(AposA8Align.CENTER, content.upper(), AposA8CommandType.PRINT_TEXT))
            report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_HASH,
                                          AposA8CommandType.PRINT_TEXT))

        if whopper_wifi_code is not None and model is not None:
            report.append(AposA8PrintLine(AposA8Align.CENTER,
                                          sysactions.translate_message(model, "WIFI_PRINTER_MESSAGE"),
                                          AposA8CommandType.PRINT_TEXT))
            report.append(AposA8PrintLine(AposA8Align.CENTER,
                                          sysactions.translate_message(model,
                                                                       "WIFI_ACCESS_CODE",
                                                                       "%s" % whopper_wifi_code),
                                          AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        # Campo de texto customizado exibido na área de informações adicionais
        if self.customer_text_field != "":
            for customer_text in self.customer_text_field.split('\\n'):
                report.append(AposA8PrintLine(AposA8Align.CENTER,
                                              customer_text.replace("\\n", ""),
                                              AposA8CommandType.PRINT_TEXT))
            report.append(
                AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        satk = insert_space(sat_fiscal_response.find("satkey").text.strip(), 4)
        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      str(satk),
                                      AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

        cpf_cnpj = sat_fiscal_response.find("CustomerCPF").text
        cpf_text = "Consumidor: %s" % ("Nao informado" if not cpf_cnpj else cpf_cnpj)
        report.append(AposA8PrintLine(AposA8Align.CENTER, cpf_text, AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      "SAT No. %s" % sat_fiscal_response.find("satno").text.strip(),
                                      AposA8CommandType.PRINT_TEXT))

        report.append(AposA8PrintLine(AposA8Align.CENTER,
                                      sat_fiscal_response.find("datetime").text.strip(),
                                      AposA8CommandType.PRINT_TEXT))

        return report



from production.box import DecreaseTimeTagBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class DecreaseTimeTagBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(DecreaseTimeTagBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> DecreaseTimeTagBox
        def get_config(config_name):
            return self.get_extra_config(config, config_name)

        return DecreaseTimeTagBox(config.name,
                                  config.sons,
                                  self.get_logger(config),
                                  get_config)

    def get_extra_config(self, config, config_name):
        config_value = self.get_extra(config, config_name, None)
        if config_value is None:
            self.loggers['DEBUG'].info("({}) does not exist".format(config_name))
            raise Exception("({}) does not exist".format(config_name))

        return config_value



from deliveryreport.model import DeliveryOrder


class DeliveryReport(object):
    def __init__(self, order_repository, product_repository):
        self.order_repository = order_repository
        self.product_repository = product_repository

    def get_delivery_order(self, order_id):
        unfiltered_order_json = self.order_repository.get_order_json(order_id)
        order_json = [x for x in unfiltered_order_json if x][0]
        return DeliveryOrder(order_json)

    def get_products(self):
        return self.product_repository.get_products()

    def get_cfh_products(self):
        return self.product_repository.get_cfh_products()



import json
import logging
from typing import Optional, List

from report.command import AlignCommand, BoldCommand, FontCommand, RepeatCommand, NewLineCommand

from ._Report import Report
from ._Part import Part
from ._JsonReport import JsonReport

logger = logging.getLogger("PosActions")


class ReportParser(object):
    def parse(self, report):
        # type: (str) -> Optional[Report]

        try:
            report_dict = json.loads(report, "utf-8")
        except Exception as _:
            logger.exception("Error parsing report JSON data: {}".format(report))
            return None

        if "key" not in report_dict or report_dict["key"] != Report.key:
            logger.error("Error getting report key")
            return None

        if "parts" not in report_dict or not isinstance(report_dict["parts"], list):
            logger.error("Error getting report parts")
            return None

        parts = []
        for part_dict in report_dict["parts"]:
            text = part_dict["text"]
            commands_json = []
            if "commands" in part_dict:
                commands_json = part_dict["commands"]
                if commands_json is None:
                    commands_json = []

            part_commands = self._get_commands(commands_json)
            parts.append(Part(text, part_commands))

        return JsonReport(parts, report_dict["width"])

    @staticmethod
    def _get_commands(
            commands_json   # type: str
    ):
        # type: (...) -> List

        part_commands = []
        for command in commands_json:
            command_type = command["type"]
            if command_type == "align":
                part_commands.append(AlignCommand(AlignCommand.Alignment(command["alignment"])))
                continue

            if command_type == "bold":
                part_commands.append(BoldCommand(command["activate"]))
                continue

            if command_type == "font":
                part_commands.append(FontCommand(command["font"]))
                continue

            if command_type == "repeat":
                part_commands.append(RepeatCommand(command["count"], command["text"]))
                continue

            if command_type == "newLine":
                part_commands.append(NewLineCommand())

        return part_commands



from enum import Enum


class TransferType(Enum):
    INITIAL_AMOUNT = '1'
    TRANSFER_SKIM = '2'
    TRANSFER_CASH_IN = '3'
    TRANSFER_CASH_OUT = '4'
    FINAL_AMOUNT = '5'
    DECLARED_AMOUNT = '6'



from mw import MwTableOrderRetriever, MsgBusTableService, RealClock
from reportsdata.tablereport import TableReportPresenter, TableReportGenerator
from reportsdata.tablereport.dto import TableReportFilterDto


class ReportGenerator(object):
    @staticmethod
    def generate(mb_context,
                 pos_id,
                 table_id,
                 print_pre_account_by="TABLE",
                 hide_unpriced_items="True",
                 pre_account_with_discount=True):

        table_order_retriever = MwTableOrderRetriever(MsgBusTableService())

        hide_unpriced_items = True if hide_unpriced_items == "True" else False
        pre_account_with_discount = True if pre_account_with_discount == "True" else False

        generator = TableReportGenerator(
                TableReportFilterDto(pos_id, table_id, None),
                RealClock(),
                table_order_retriever
        )

        presenter = TableReportPresenter(print_pre_account_by, hide_unpriced_items, pre_account_with_discount, pos_id)

        return generator, presenter



from dailygoals_services.model import SaleChallenge
from dailygoals_services.model.repository import DailyGoalsRepository


class UpdateSalesChallengeInteractor:
    def __init__(self, daily_goals_repository):
        # type: (DailyGoalsRepository) -> None

        self.daily_goals_repository = daily_goals_repository

    def update_sales_challenge(self, challenge):
        # type: (SaleChallenge) -> int

        return self.daily_goals_repository.update_sales_challenge(challenge)



"""
SignXML utility functions

bytes_to_long, long_to_bytes copied from https://github.com/dlitz/pycrypto/blob/master/lib/Crypto/Util/number.py
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os, sys, re, struct, textwrap
from xml.etree import ElementTree as stdlibElementTree
from base64 import b64encode, b64decode

from eight import str, bytes
from lxml import etree
from defusedxml.lxml import fromstring
from pyasn1.type import univ

from ..exceptions import RedundantCert, InvalidCertificate

USING_PYTHON2 = True if sys.version_info < (3, 0) else False

PEM_HEADER = "-----BEGIN CERTIFICATE-----"
PEM_FOOTER = "-----END CERTIFICATE-----"

def ensure_bytes(x, encoding="utf-8", none_ok=False):
    if none_ok is True and x is None:
        return x
    if not isinstance(x, bytes):
        x = x.encode(encoding)
    return x

def ensure_str(x, encoding="utf-8", none_ok=False):
    if none_ok is True and x is None:
        return x
    if not isinstance(x, str):
        x = x.decode(encoding)
    return x

def bytes_to_long(s):
    """bytes_to_long(string) : long
    Convert a byte string to a long integer.

    This is (essentially) the inverse of long_to_bytes().
    """
    if isinstance(s, int):
        # On Python 2, indexing into a bytearray returns a byte string; on Python 3, an int.
        return s
    acc = 0
    if USING_PYTHON2:
        acc = long(acc)  # noqa
    unpack = struct.unpack
    length = len(s)
    if length % 4:
        extra = (4 - length % 4)
        s = b'\000' * extra + s
        length = length + extra
    for i in range(0, length, 4):
        acc = (acc << 32) + unpack(b'>I', s[i:i+4])[0]
    return acc

def long_to_bytes(n, blocksize=0):
    """long_to_bytes(n:long, blocksize:int) : string
    Convert a long integer to a byte string.

    If optional blocksize is given and greater than zero, pad the front of the
    byte string with binary zeros so that the length is a multiple of
    blocksize.
    """
    # after much testing, this algorithm was deemed to be the fastest
    s = b''
    if USING_PYTHON2:
        n = long(n)  # noqa
    pack = struct.pack
    while n > 0:
        s = pack(b'>I', n & 0xffffffff) + s
        n = n >> 32
    # strip off leading zeros
    for i in range(len(s)):
        if s[i] != b'\000'[0]:
            break
    else:
        # only happens when n == 0
        s = b'\000'
        i = 0
    s = s[i:]
    # add back some pad bytes.  this could be done more efficiently w.r.t. the
    # de-padding being done above, but sigh...
    if blocksize > 0 and len(s) % blocksize:
        s = (blocksize - len(s) % blocksize) * b'\000' + s
    return s

pem_regexp = re.compile("{header}\n(.+?){footer}".format(header=PEM_HEADER, footer=PEM_FOOTER), flags=re.S)

def strip_pem_header(cert):
    try:
        return re.search(pem_regexp, ensure_str(cert)).group(1)
    except Exception:
        return ensure_str(cert)

def add_pem_header(bare_base64_cert):
    bare_base64_cert = ensure_str(bare_base64_cert)
    if bare_base64_cert.startswith(PEM_HEADER):
        return bare_base64_cert
    return PEM_HEADER + "\n" + textwrap.fill(bare_base64_cert, 64) + "\n" + PEM_FOOTER

def iterate_pem(certs):
    for match in re.findall(pem_regexp, ensure_str(certs)):
        yield match

class DERSequenceOfIntegers(univ.SequenceOf):
    componentType = univ.Integer()
    def __init__(self, integers):
        univ.SequenceOf.__init__(self)
        for pos, i in enumerate(integers):
            self.setComponentByPosition(pos, i)

class Namespace(dict):
    __getattr__ = dict.__getitem__

class XMLProcessor:
    _schema, _default_parser = None, None

    @classmethod
    def schema(cls):
        if cls._schema is None:
            schema_path = os.path.join(os.path.dirname(__file__), "..", "schemas", cls.schema_file)
            cls._schema = etree.XMLSchema(etree.parse(schema_path))
        return cls._schema

    @property
    def parser(self):
        if self._parser is None:
            if self._default_parser is None:
                self._default_parser = etree.XMLParser()
            return self._default_parser
        return self._parser

    def get_root(self, data):
        if isinstance(data, (str, bytes)):
            return fromstring(data, parser=self.parser)
        elif isinstance(data, stdlibElementTree.Element):
            # TODO: add debug level logging statement re: performance impact here
            return fromstring(stdlibElementTree.tostring(data, encoding="utf-8"))
        else:
            # HACK: deep copy won't keep root's namespaces resulting in an invalid digest
            # We use a copy so we can modify the tree
            # TODO: turn this off for xmlenc
            return fromstring(etree.tostring(data))

def hmac_sha1(key, message):
    from cryptography.hazmat.primitives import hashes, hmac
    from cryptography.hazmat.backends import default_backend
    hasher = hmac.HMAC(key, hashes.SHA1(), backend=default_backend())
    hasher.update(message)
    return hasher.finalize()

def raw_p_sha1(secret, seed, sizes=()):
    """
    Derive one or more keys from secret and seed.
    (See specs part 6, 6.7.5 and RFC 2246 - TLS v1.0)
    Lengths of keys will match sizes argument

    Source: https://github.com/FreeOpcUa/python-opcua
    key_sizes = (signature_key_size, symmetric_key_size, 16)
    (sigkey, key, init_vec) = p_sha1(nonce2, nonce1, key_sizes)
    """
    full_size = 0
    for size in sizes:
        full_size += size

    result = b''
    accum = seed
    while len(result) < full_size:
        accum = hmac_sha1(secret, accum)
        result += hmac_sha1(secret, accum + seed)

    parts = []
    for size in sizes:
        parts.append(result[:size])
        result = result[size:]
    return tuple(parts)

def p_sha1(client_b64_bytes, server_b64_bytes):
    client_bytes, server_bytes = b64decode(client_b64_bytes), b64decode(server_b64_bytes)
    return b64encode(raw_p_sha1(client_bytes, server_bytes, (len(client_bytes), len(server_bytes)))[0]).decode()


def _add_cert_to_store(store, cert):
    from OpenSSL.crypto import X509StoreContext, X509StoreContextError, Error as OpenSSLCryptoError
    try:
        X509StoreContext(store, cert).verify_certificate()
    except X509StoreContextError as e:
        raise InvalidCertificate(e)
    try:
        store.add_cert(cert)
        return cert
    except OpenSSLCryptoError as e:
        if e.args == ([('x509 certificate routines', 'X509_STORE_add_cert', 'cert already in hash table')],):
            raise RedundantCert(e)
        raise

def verify_x509_cert_chain(cert_chain, ca_pem_file=None, ca_path=None):
    """
    Look at certs in the cert chain and add them to the store one by one.
    Return the cert at the end of the chain. That is the cert to be used by the caller for verifying.
    From https://www.w3.org/TR/xmldsig-core2/#sec-X509Data:
    "All certificates appearing in an X509Data element must relate to the validation key by either containing it
    or being part of a certification chain that terminates in a certificate containing the validation key.
    No ordering is implied by the above constraints"
    """
    from OpenSSL import SSL
    context = SSL.Context(SSL.TLSv1_METHOD)
    if ca_pem_file is None and ca_path is None:
        import certifi
        ca_pem_file = certifi.where()
    context.load_verify_locations(ensure_bytes(ca_pem_file, none_ok=True), capath=ca_path)
    store = context.get_cert_store()
    certs = list(reversed(cert_chain))
    end_of_chain, last_error = None, None
    while len(certs) > 0:
        for cert in certs:
            try:
                end_of_chain = _add_cert_to_store(store, cert)
                certs.remove(cert)
                break
            except RedundantCert:
                certs.remove(cert)
                if end_of_chain is None:
                    end_of_chain = cert
                break
            except Exception as e:
                last_error = e
        else:
            raise last_error
    return end_of_chain



from commons.dto import DefaultHeaderDto  # noqa
from _TransfersByDayDto import TransfersByDayDto  # noqa


class CashTransferReportDto(object):
    def __init__(self, header_transfer_dto, all_transfers_dto, pos_id):
        # type: (DefaultHeaderDto, [TransfersByDayDto]) -> None
        self.header_transfer_dto = header_transfer_dto
        self.all_transfers_dto = all_transfers_dto
        self.pos_id = pos_id



from commons.dto import ReportColumnDefinition, AlignTypes
from commons.util import TableReport
from typing import List, Union  # noqa
from reports_app.cashreport2.dto import CashReportDto  # noqa
from report import Part, Presenter, SimpleReport
from report import PartGenerator  # noqa
from report.command import RepeatCommand, AlignCommand


class PeriodCashReportPresenter(Presenter):
    def present(self, dto):
        report = [
            Part(commands=[RepeatCommand(38, "=")]), Part("\n"),
            Part("$PERIOD_CASH_REPORT_{}".format(dto.header.report_type),
                 [AlignCommand(AlignCommand.Alignment.center)]), Part("\n")
        ]

        column_definitions = [
            ReportColumnDefinition(width=17,
                                   fill_with_char=".",
                                   before_text="  ",
                                   after_text="",
                                   after_fill_text=":",
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(width=22,
                                   fill_with_char=" ",
                                   before_text=" ",
                                   align=AlignTypes.LEFT)
        ]

        pos_label = "$ALL"
        if dto.header.pos is not None:
            pos_label = str(dto.header.pos)

        operator_label = "$ALL"
        if dto.header.operator_id is not None:
            operator_label = "{} - {}".format(dto.header.operator_id, dto.header.operator_name)

        lines = [
            ["$STORE_POS", "{} / {}".format(dto.header.store_code, pos_label)],
            ["$OPERATOR", operator_label],
            ["$PERIOD", "#DATE({})-#DATE({})"
                .format(dto.header.initial_date.isoformat(), dto.header.end_date.isoformat())],
            ["$DATE", "#DATETIME({})".format(dto.header.generated_time.isoformat())]
        ]
        header = TableReport(lines=lines, column_definitions=column_definitions)
        report.append(header)

        report.extend([
            Part(commands=[RepeatCommand(38, "=")]), Part("\n")
        ])

        column_definitions = [
            ReportColumnDefinition(width=18,
                                   fill_with_char=" ",
                                   before_text="",
                                   after_text="",
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(width=8,
                                   fill_with_char=" ",
                                   before_text="",
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(width=4,
                                   fill_with_char=" ",
                                   before_text=" ",
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(width=8,
                                   fill_with_char=" ",
                                   before_text=" ",
                                   align=AlignTypes.LEFT)
        ]

        lines = [["$DESCRIPTION", "$QTY", "", "$TOTAL"]]
        body_header = TableReport(lines=lines, column_definitions=column_definitions)

        report.append(body_header)

        column_definitions = [
            ReportColumnDefinition(width=18,
                                   fill_with_char=".",
                                   before_text="",
                                   after_text="",
                                   after_fill_text=":",
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(width=8,
                                   fill_with_char=" ",
                                   before_text="[",
                                   after_text="]",
                                   align=AlignTypes.RIGHT),
            ReportColumnDefinition(width=4,
                                   fill_with_char=" ",
                                   before_text="",
                                   after_text="",
                                   align=AlignTypes.RIGHT),
            ReportColumnDefinition(width=8,
                                   fill_with_char=" ",
                                   before_text="",
                                   after_text="",
                                   align=AlignTypes.RIGHT)
        ]

        lines = [
            self.create_report_line("$INITIAL_BALANCE", dto.body.initial_float),
            ["$OPERATORS", str(dto.body.operator_count), "", ""],
            self.create_report_line("$TOTAL_SALES", dto.body.total_sales),
            self.create_report_line("$VOIDS", dto.body.voids),
            self.create_report_line("$NET_SALES", dto.body.net_sales)
        ]

        for tender in dto.body.tender_breakdown:
            lines.append(self.create_report_line("  $TENDER_{}".format(tender.detail), tender))

        lines.extend([
            self.create_report_line("$CASH_OUTS", dto.body.cash_outs),
            self.create_report_line("$CASH_INS", dto.body.cash_ins),
            ["$VALUE_IN_DRAWER", "", "#CURRENCY_SYMBOL()", "#NUMBER({0:.6f})".format(dto.body.value_in_drawer)]
        ])

        report.append(TableReport(lines=lines, column_definitions=column_definitions))

        report.append(Part(commands=[RepeatCommand(38, "=")]))
        return SimpleReport(report)

    def create_report_line(self, label, report_line):
        return [label,
                str(report_line.quantity),
                "#CURRENCY_SYMBOL()",
                "#NUMBER({0:.6f})".format(report_line.value)]



from typing import Callable

from nperipherals_core.model.device import DeviceServiceThread


class HealthCheckInteractor(object):
    def __init__(self, device_service_thread, destroy_func, build_func):
        # type: (DeviceServiceThread, Callable[[], None], Callable[[], None]) -> None

        self.device_service_thread = device_service_thread
        self.destroy_func = destroy_func
        self.build_func = build_func

    def restart_dead_threads(self):
        # type: () -> None

        if self.device_service_thread is not None and not self.device_service_thread.running:
            self.destroy_func()
            self.build_func()



class InvalidRequestException(Exception):
    pass


class CustomerExistsException(Exception):
    pass


class CustomerNonExistsException(Exception):
    pass


class InsertCustomerException(Exception):
    pass


class UpdateCustomerException(Exception):
    pass



from typing import List

from mw_helper import ensure_iterable

from production.model import ProductionOrder
from ._OrderCondition import OrderCondition


class SaleTypeCondition(OrderCondition):
    def __init__(self, sale_types):
        # type: (List[str]) -> None

        self.valid_sale_types = ensure_iterable(sale_types)
        self.valid_sale_types = [x.lower() for x in self.valid_sale_types]

    def evaluate(self, order):
        # type: (ProductionOrder) -> bool

        return order.sale_type.lower() in self.valid_sale_types



from typing import Union


class TableReportSaleLineDto(object):

    def __init__(self):
        self.product_name = ""  # type: str
        self.price = 0.0        # type: float
        self.qty = 0            # type: float
        self.default_qty = 0    # type: Union[int, None]
        self.item_type = ""     # type: str
        self.seat = None        # type: Union[int, None]
        self.level = 0          # type: int
        self.comment = ""       # type: Union[str, None]
        self.line_number = 0    # type: int
        self.service_period = None # type: Union[str, None]



import json
import os
import socket
from logging import getLogger

from externals.device import NetworkDevice
from nperipherals_core.model.custom_exceptions import DeviceNotConnected, PrinterDoorOpened, PrinterNotConnected
from nperipherals_core.model.device import WriterDevice, CashDrawer
from nperipherals_core.model.device_status import CashDrawerStatus, DeviceStatus
from typing import Dict, List

logger = getLogger(__name__)


class NetworkEscPosPrinterDevice(WriterDevice, NetworkDevice, CashDrawer):
    REQUEST_STATUS_CODE = b'\x10\x04\x01'
    SOCKET_BUF_SIZE = 1

    def __init__(self, device_configurations):
        # type: (Dict[unicode, unicode]) -> None

        super(NetworkEscPosPrinterDevice, self).__init__(logger, device_configurations)
        self.json_data = self.get_json_data(device_configurations.get("TagsFile", ""))  # type: Dict[str,str]
        self.pre_commands = self.parse_commands(device_configurations.get("PrePrintCmds", []))  # type: str
        self.post_commands = self.parse_commands(device_configurations.get("PostPrintCmds", []))  # type: str
        self.write_retries = int(device_configurations.get("Retries", 1))  # type: int
        self.write_retry_sleep_in_seconds = float(device_configurations.get("RetriesSleepSeconds", 1.0))  # type: float

    @staticmethod
    def get_json_data(tags_path):
        # type: (str) -> Dict[str,str]
        data = {}
        if os.path.exists(tags_path):
            with open(tags_path, 'rb') as f:
                data = json.load(f)
        return data

    def parse_commands(self, commands):
        # type: (List[str]) -> str

        parsed = b""
        for command in commands:
            parsed += self.json_data.get(command, command)
        return bytes(parsed)

    def _get_bin_status(self):
        # type: () -> str

        self.socket.send(self.REQUEST_STATUS_CODE)
        status = self.socket.recv(self.SOCKET_BUF_SIZE)
        return format(ord(status), '08b')

    def get_status(self):
        # type: () -> DeviceStatus

        bin_status = self._get_bin_status()

        if "1" == bin_status[-6]:
            raise PrinterDoorOpened()

        if "1" == bin_status[-4]:
            raise PrinterNotConnected()

        return DeviceStatus.Ok

    def get_drawer_status(self):
        # type: () -> CashDrawerStatus

        try:
            bin_status = self._get_bin_status()

            if "0" == bin_status[-3]:
                return CashDrawerStatus.Opened

            return CashDrawerStatus.Closed

        except (socket.error, AttributeError):
            raise DeviceNotConnected()

    def write(self, data):
        # type: (unicode) -> None

        try:
            if self.pre_commands:
                self.socket.sendall(self.pre_commands)

            self.socket.sendall(data)

            if self.post_commands:
                self.socket.sendall(self.post_commands)

        except (socket.error, AttributeError):
            raise DeviceNotConnected()

    def open_drawer(self):
        # type: () -> None

        try:
            self._send_open_drawer_command()
        except (socket.error, AttributeError) as _:
            raise DeviceNotConnected()

    def _send_open_drawer_command(self):
        cash_drawer_pin_2_pulse = b'\x1bp\x0022'
        cash_drawer_pin_5_pulse = b'\x1bp\x0122'

        self.socket.sendall(cash_drawer_pin_2_pulse)
        self.socket.sendall(cash_drawer_pin_5_pulse)

    def get_write_retries(self):
        # type: () -> int

        return self.write_retries

    def get_write_retry_sleep_in_seconds(self):
        # type: () -> float

        return self.write_retry_sleep_in_seconds



import logging

import bustoken
from messagehandler import EventHandler
from msgbus import MBEasyContext, MBMessage, TK_SYS_ACK, TK_SYS_NAK
from typing import Optional

from fingerprintreader import DigitalPersonaFingerPrintReader, DigitalPersonaFingerPrintException

logger = logging.getLogger("FingerPrintReader")


class FingerPrintReaderEventHandler(EventHandler):
    def __init__(self, mbcontext, finger_print_reader):
        # type: (MBEasyContext, DigitalPersonaFingerPrintReader) -> None
        super(FingerPrintReaderEventHandler, self).__init__(mbcontext)

        self.finger_print_reader = finger_print_reader
        self.token_dictionary = {
            bustoken.TK_FPR_ENROLL_USER: self._handle_enroll_user,
            bustoken.TK_FPR_IDENTIFY_USER: self._handle_identify_user,
            bustoken.TK_FPR_AUTHORIZE_USER: self._handle_authorize_user,
            bustoken.TK_FPR_OK: self._handle_service_ok,
            bustoken.TK_FPR_RE_ENROLL_USER: self._handle_re_enroll_user
        }

    def get_handled_tokens(self):
        return self.token_dictionary.keys()

    def handle_message(self, msg):
        # type: (MBMessage) -> None
        event_handler = self.token_dictionary[msg.token]
        event_handler(self.mbcontext, msg)

    def handle_event(self, subject, evt_type, data, msg):
        raise NotImplementedError()

    def terminate_event(self):
        pass

    def _handle_identify_user(self, mbcontext, msg, authorization=False):
        # type: (MBEasyContext, MBMessage, Optional[bool]) -> None
        try:
            pos_id = msg.data
            finger_print_user_id = self.finger_print_reader.authenticate(pos_id, authorization)
            msg.token = TK_SYS_ACK
            logger.info("Successfully authentication for POS: #{}".format(pos_id))
            mbcontext.MB_ReplyMessage(msg, data=finger_print_user_id)
        except DigitalPersonaFingerPrintException as ex:
            data = None
            if ex.error_code == -2: # Timeout
                msg.token = bustoken.TK_FPR_TIMEOUT
            else:
                msg.token = TK_SYS_NAK
                data = str(ex.error_code) + " " + ex.message
                logger.error(data)

            mbcontext.MB_ReplyMessage(msg, data=data)
        except Exception as e:
            msg.token = TK_SYS_NAK
            data = e.message
            logger.exception(data)
            mbcontext.MB_ReplyMessage(msg, data=data)

    def _handle_authorize_user(self, mbcontext, msg):
        # type: (MBEasyContext, MBMessage) -> None
        try:
            self._handle_identify_user(mbcontext, msg, True)
        except Exception as _:
            logger.exception("[_handle_authorize_user]")

    def _handle_enroll_user(self, mbcontext, msg):
        # type: (MBEasyContext, MBMessage) -> None
        try:
            self._handle_enroll_and_re_enroll_user(mbcontext, msg, False)
        except Exception as _:
            logger.exception("[_handle_enroll_user]")

    def _handle_re_enroll_user(self, mbcontext, msg):
        # type: (MBEasyContext, MBMessage) -> None
        try:
            self._handle_enroll_and_re_enroll_user(mbcontext, msg, True)
        except Exception as _:
            logger.exception("[_handle_re_enroll_user]")

    def _handle_enroll_and_re_enroll_user(self, mbcontext, msg, replace):
        # type: (MBEasyContext, MBMessage, Optional[bool]) -> None
        user_id, pos_id = msg.data.split(";")
        if user_id is None:
            msg.token = TK_SYS_NAK
            data = "No user_id received"
            logger.info(data)
            mbcontext.MB_ReplyMessage(msg, data=data)
            return

        try:
            fmd_data = self.finger_print_reader.enroll_user(user_id, pos_id, replace)
            msg.token = TK_SYS_ACK
            mbcontext.MB_ReplyMessage(msg, data=fmd_data)
        except DigitalPersonaFingerPrintException as ex:
            msg.token = TK_SYS_NAK
            data = str(ex.error_code) + " " + ex.message

            if ex.error_code < 0:
                if ex.error_code == -1:
                    msg.token = bustoken.TK_FPR_FINGER_PRINT_ALREADY_REGISTERED
                    data = "Impressao digital ja cadastrada"
                    logger.error(data)

            mbcontext.MB_ReplyMessage(msg, data=data)
        except Exception as e:
            msg.token = TK_SYS_NAK
            data = e.message
            logger.exception(data)
            mbcontext.MB_ReplyMessage(msg, data=data)

    def _handle_service_ok(self, mbcontext, msg):
        # type: (MBEasyContext, MBMessage) -> None
        data = None
        try:
            self.finger_print_reader.check_device()
            logger.info("Device is found")
            msg.token = TK_SYS_ACK
        except Exception as e:
            logger.exception("Device is not found")
            data = e.message
            msg.token = TK_SYS_NAK

        mbcontext.MB_ReplyMessage(msg, data=data)


class FingerPrintReaderCallback:
    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        self.mbcontext = mbcontext

    def finger_print_reader_callback(self, event, pos_id, extra_data=""):
        # type: (str, str, Optional[str]) -> None
        data = pos_id
        if extra_data:
            data = pos_id + ";" + str(extra_data)

        logger.info("Callback: {}".format(data))
        self.mbcontext.MB_EasyEvtSend("FPR_CALLBACK", event, data)



# -*- coding: utf-8 -*-

from models import NfeCardBrand


class PosTenderBrandToNfeCardBrand(object):
    @staticmethod
    def convert(pos_tender_brand):
        # type: (unicode) -> str

        tender_brand_map = {

        }

        nfe_card_brand = tender_brand_map.get(pos_tender_brand, NfeCardBrand.outros)
        return str(nfe_card_brand)



from iso8601 import iso8601
from typing import Dict

from application.model.benefit._RequiredItemOnSale import RequiredItemOnSale
from application.model.benefit._RequiredGroupOnSale import RequiredGroupOnSale
from application.model.benefit._ValidHours import ValidHours


class Conditions(object):
    def __init__(self, obj):
        # type: (Dict) -> None

        self.user_min_level = obj.get("userMinLevel")
        self.unique_for_sale = bool(obj.get("uniqueForSale"))
        self.ignore_burned_items = bool(obj.get("ignoreBurnedItems"))
        self.single_use = bool(obj.get("singleUse")) or self.ignore_burned_items
        self.burn_items = bool(obj.get("burnItems", True))
        self.pod_type = list(obj.get("podType")) if obj.get("podType") else None
        self.min_sale_amount = float(obj.get("minSaleAmount")) if obj.get("minSaleAmount") else None
        self.max_sale_amount = float(obj.get("maxSaleAmount")) if obj.get("maxSaleAmount") else None
        start_period = obj.get("validStartPeriod")
        self.valid_start_period = iso8601.parse_date(start_period) if start_period else None
        valid_end_period = obj.get("validEndPeriod")
        self.valid_end_period = iso8601.parse_date(valid_end_period) if valid_end_period else None
        valid_hours = obj.get("validHours")
        start_hour = valid_hours.get("startHour") if valid_hours else None
        end_hour = valid_hours.get("endHour") if valid_hours else None
        self.valid_hours = ValidHours(start_hour, end_hour)
        self.valid_days = list(obj.get("validDays")) if obj.get("validDays") else None
        self.required_items_on_sale = [RequiredItemOnSale(x) for x in obj.get("requiredItemsOnSale", [])]
        self.service_type = list(obj.get("serviceType")) if obj.get("serviceType") else None
        self.sale_type = list(obj.get("saleType")) if obj.get("saleType") else None
        self.sector = list(obj.get("sector")) if obj.get("sector") else None
        self.min_seats = obj.get("minSeats")
        self.discount_max_value = obj.get("discountMaxValue")
        self.service_period = list(obj.get("servicePeriods")) if obj.get("servicePeriods") else None
        self.brands = list(obj.get("brands")) if obj.get("brands") else None
        self.required_groups_on_sale = [RequiredGroupOnSale(x) for x in obj.get("requiredGroupsOnSale", [])]



class DeviceNotConnected(Exception):
    pass


class UnknownDevice(Exception):
    pass


class UnmappedDeviceServiceThread(Exception):
    pass


class FailedWritingToDevice(Exception):
    pass


class PrinterDoorOpened(Exception):
    pass


class PrinterNotConnected(Exception):
    pass



import json
import logging
import logging.config
import logging.handlers
import os
import sys

try:
    from StringIO import StringIO  ## for Python 2
except ImportError:
    from io import StringIO  ## for Python 3

from datetime import datetime
from decimal import Decimal, ROUND_HALF_UP
from threading import Lock
from time import sleep
from xml.etree import cElementTree as eTree

import tzlocal
from enum import Enum
from typing import Union, Dict
from unidecode import unidecode


class IsoFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, style='%'):
        super(IsoFormatter, self).__init__(fmt, datefmt)
        self.lock = Lock()
        self.count = 0
        self.max_count = 99

    def get_count(self):
        with self.lock:
            self.count += 1
            if self.count > self.max_count:
                self.count = 1
            return self.count

    def formatTime(self, record, datefmt=None):
        value = datetime.fromtimestamp(record.created, tzlocal.get_localzone())
        base_date = datetime.strftime(value, "%Y-%m-%dT%H:%M:%S")
        base_date += "." + str(int(record.msecs)).zfill(3)
        base_date += datetime.strftime(value, "%z")
        return base_date

    def format(self, record):
        if not hasattr(record, "count"):
            record.count = str(self.get_count()).zfill(2)
        return super(IsoFormatter, self).format(record)


def config_logger(loader_path,
                  log_name='component',
                  configure_root=False,
                  root_level=logging.WARN,
                  max_files=5,
                  max_size=5242880,
                  add_error_handler=False):
    real_path = os.path.realpath(loader_path)
    if os.path.isdir(real_path):
        dir_path = real_path
    else:
        dir_path = os.path.dirname(real_path)

    filename = os.path.join(dir_path, log_name + '.log')
    filename_error = os.path.join(dir_path, log_name + '.error.log')

    formatter = IsoFormatter("%(asctime)-6s: %(count)2s - %(name)s - %(levelname)s - "
                             "%(thread)d - %(module)s:%(filename)s:%(lineno)s:%(funcName)s - "
                             "%(message)s")

    comp_logger = logging.getLogger(log_name)
    comp_logger.setLevel(logging.DEBUG)
    comp_logger.propagate = False

    rotating_file_logger = logging.handlers.RotatingFileHandler(filename, 'a', max_size, max_files, encoding="utf-8")
    rotating_file_logger.setLevel(logging.DEBUG)
    rotating_file_logger.setFormatter(formatter)
    comp_logger.addHandler(rotating_file_logger)

    if add_error_handler:
        rotating_file_logger_error = logging.handlers.RotatingFileHandler(filename_error,
                                                                          'a',
                                                                          max_size,
                                                                          max_files,
                                                                          encoding="utf-8")
        rotating_file_logger_error.setLevel(logging.ERROR)
        rotating_file_logger_error.setFormatter(formatter)
        comp_logger.addHandler(rotating_file_logger_error)

    if configure_root:
        root_logger = logging.getLogger()
        root_logger.addHandler(rotating_file_logger)
        root_logger.setLevel(root_level)


def import_pydevd(loader_path, port, suspend=False):
    try:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        debug_lib_dir = os.path.join(current_dir, "..", "debug_lib")
        if os.path.exists(debug_lib_dir):
            sys.path.append(debug_lib_dir)
            from debughelper import import_pydevd as real_import_pydevd
            real_import_pydevd(loader_path, port, suspend)
    except ImportError:
        return


def get_path(directory='', file_name_with_extension=''):
    # type: (str, str) -> str
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), directory, file_name_with_extension)


def read_json_from_file_as_dictionary(json_file_path):
    with open(json_file_path, 'rb') as content_file:
        complete_json = content_file.read()
        return json.loads(complete_json.decode("utf-8"))


def remove_accents(text):
    # type: (str) -> str
    if sys.version_info[0] == 2:
        text = unicode(text)
    else:
        text = str(text)

    return unidecode(text)


def json_serialize(dictionary):
    # type: (Dict) -> str
    try:
        return json.dumps(dictionary, default=lambda o: getattr(o, '__dict__', str(o)))
    except ValueError as _:
        return dictionary


def json_pretty_format(dictionary):
    # type: (Dict) -> str

    return json.dumps(dictionary, sort_keys=True, indent=1, default=lambda o: getattr(o, '__dict__', str(o)))


def retry(logger, method_name, n, sleep_seconds, trying_to_do, is_success, success, failed, on_exception=None):
    for attempt in range(1, n + 1):
        # noinspection PyBroadException
        try:
            logger.debug("Retry [{}] :: trying_to_do {}/{}".format(method_name, attempt, n))
            response = trying_to_do()
            if is_success(response):
                logger.debug("Retry [{}] :: is_success {}/{}".format(method_name, attempt, n))
                return success(response)
            else:
                logger.debug("Retry [{}] :: handle_error {}/{}".format(method_name, attempt, n))
                failed(response)
        except Exception as _:
            logger.exception("Retry [{}] :: EXCEPTION {}/{}".format(method_name, attempt, n))
            if on_exception:
                on_exception()

        sleep(sleep_seconds)

    logger.error("Retry {} :: FAILED after {} tentatives".format(method_name, n))


def dict_to_object(dictionary, obj):
    # type: (dict, object) -> None

    for key in dictionary:
        if hasattr(obj, key):
            setattr(obj, key, dictionary[key])


def build_dict(seq, key):
    return dict((d[key], dict(d, index=index)) for (index, d) in enumerate(seq))


def get_class_by_name(kls):
    # type: (unicode) -> type
    parts = kls.split(".")
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m


def remove_xml_namespace(xml):
    # type: (Union[str, unicode]) -> eTree.ElementTree
    it = eTree.iterparse(StringIO(xml))
    for _, el in it:
        if '}' in el.tag:
            el.tag = el.tag.split('}', 1)[1]
    return it.root


def round_half_away_from_zero(number, precision=0):
    number = Decimal(number)
    value = number.quantize(Decimal('10') ** -(precision + 2))
    return float(value.quantize(Decimal('10') ** -precision, rounding=ROUND_HALF_UP))


class ExtendedEnum(Enum):

    @classmethod
    def list_values(cls):
        # type: () -> list
        return list(map(lambda enum: enum.value, cls))

    @classmethod
    def list_names(cls):
        # type: () -> list
        return list(map(lambda enum: enum.name, cls))

    @classmethod
    def list_enums_tuple(cls):
        # type: () -> list
        return list(map(lambda enum: (enum.name, enum.value), cls))

    @classmethod
    def get_name(cls, value):
        # type: (int) -> Union[str, None]
        if value not in cls.list_values():
            return None

        return next(str(enum.name) for enum in cls if enum.value == value)

    @classmethod
    def get_value(cls, name):
        # type: (str) -> Union[int, None]
        if name not in cls.list_names():
            return None

        return next(int(enum.value) for enum in cls if enum.name == name)



from ._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates


class ComboServedWhenAllServedBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(ComboServedWhenAllServedBox, self).__init__(name, sons, logger)

    def change_order(self, order):
        for item in order.items:
            if item.item_type == "COMBO" and self.check_sons_tags(item.items, ["served", "no-cook"]):
                item.add_tag("served")
                order.prod_state = ProdStates.SERVED

        return order

    @staticmethod
    def check_sons_tags(sons, tags_list):
        for son in sons:
            if not son.all_has_at_least_one_tag(tags_list):
                return False
        return True



from logging import Logger

from typing import List


class Configurations(object):
    def __init__(self,
                 required_services=None,
                 service_name=None,
                 persistence_name=None,
                 logger=None,
                 supported_benefit_appliers=None,
                 retry_frequency=None,
                 max_retry_number=None,
                 retry_min_time_to_try_again=None,
                 benefit_applier_tender_ids=None):
        # type: (str, str, str, Logger, List[str], int, int, int, dict) -> None

        self.required_services = required_services
        self.service_name = service_name
        self.persistence_name = persistence_name
        self.logger = logger
        self.supported_benefit_appliers = supported_benefit_appliers
        self.retry_frequency = retry_frequency
        self.max_retry_quantity = max_retry_number
        self.retry_min_time_to_try_again = retry_min_time_to_try_again
        self.benefit_applier_tender_ids = benefit_applier_tender_ids



class AddCustomerRequest:
    def __init__(self, phone, document, name, zip_code, city, street, neighborhood, number, complement,
                 reference_point, address_id=""):
        # type: (str, str, str, str, str, str, str, str, str, str, str) -> None
        
        self.phone = phone
        self.document = document
        self.name = name
        self.zip_code = zip_code
        self.city = city
        self.street = street
        self.neighborhood = neighborhood
        self.number = number
        self.complement = complement
        self.reference_point = reference_point
        self.address_id = address_id



import json
from datetime import datetime

from domain import TableOrderRetriever, TableOrderSaleLine, TableService, TableOrder, Table
from xml.etree import cElementTree as eTree
from sysactions import get_user_information
from old_helper import round_half_away_from_zero

from typing import List, Dict, Union  # noqa


class MwTableOrderRetriever(TableOrderRetriever):
    def __init__(self, table_service):
        # type: (TableService) -> None
        self.table_service = table_service

    def get_table_order(self, pos_id, table_id):
        # type: (str, str) -> Table

        xml_table_picture = self.table_service.get_table_picture(pos_id, table_id)

        xml = eTree.XML(xml_table_picture)
        table = Table()

        table_xml = xml.find(".")
        table.table_id = table_xml.attrib["id"]
        table.total_amount = float(self._get_table_attribute_value(table_xml.attrib, "totalAmount", 0))
        table.service_seats = int(self._get_table_attribute_value(table_xml.attrib, "serviceSeats", 0))
        table.total_per_seat = float(self._get_table_attribute_value(table_xml.attrib, "totalPerSeat", 0))
        table.total_discounts = float(self._get_table_attribute_value(table_xml.attrib, "discountAmount", 0))
        tip_rate = self.table_service.get_tip_rate()
        default_tip_amount = (table.total_amount * tip_rate) / 100 if tip_rate else 0.0
        table.tip = float(table_xml.get("tip") if table_xml.get("tip") is not None else default_tip_amount)
        start_date = self._get_table_attribute_value(table_xml.attrib, "startTS", None)
        table.start_date = datetime.strptime(start_date, "%Y-%m-%d %H:%M:%S") if start_date else None
        table.user_id = table_xml.attrib["userId"] if "userId" in table_xml.attrib else ""
        user_information = eTree.XML(get_user_information(table.user_id)).find("user")
        table.user_name = user_information.get("LongName")
        self._get_table_name(table, table_xml)

        table_order_total_gross = 0.0
        total_discount_amount = 0.0
        total_tip = 0.0

        discount_by_seat = {}
        payment_by_seat = {}
        payment_by_benefit_key = {}
        tip_excluded_by_seat = {}
        used_benefit_keys = []
        discounts_tip_excluded = []
        for order_xml in xml.findall("Orders/Order"):
            total_order_amount = float(order_xml.get("totalAmount"))
            total_order_after_discount = order_xml.get("totalAfterDiscount")
            order_discount_amount = order_xml.get("discountAmount")
            order_state = order_xml.get("state")
            default_order_tip_amount = (float(order_xml.get("totalGross")) * tip_rate) / 100 if tip_rate else 0.0
            if order_state != "TOTALED":
                tip = default_order_tip_amount
            else:
                tip = float(order_xml.get("tip"))
            total_tip += tip
            table_order_total_gross += float(order_xml.get("totalGross"))
            total_discount_amount += float(order_discount_amount)
            total_order_tender = order_xml.get("totalTender")
            order_state = order_xml.get("state")
            order_id = order_xml.get("orderId")
            custom_order_properties = order_xml.findall("CustomOrderProperties/OrderProperty")
            order_tenders = order_xml.findall("TenderHistory/Tender")

            split_key = "1"
            split_key_found = False
            item_value_discounts = {}
            for custom_property in custom_order_properties:
                properties = custom_property.find(".")
                if "key" in properties.attrib:
                    if not split_key_found and properties.attrib["key"] == "SPLIT_KEY":
                        split_key = properties.attrib["value"]
                        split_key_found = True
                    if properties.attrib["key"] == "BENEFIT_LIST":
                        benefit_list = json.loads(properties.attrib["value"])
                        item_value_discounts.update(benefit_list.get("discountItemValues", {}))
                        discounts_tip_excluded.extend(benefit_list.get("discountIdTipExcluded", []))
                        self._get_promotional_tenders(benefit_list, table, payment_by_benefit_key, order_tenders)

            seats = table.service_seats
            order_sales = order_xml.findall("SaleLine")
            sale_lines = self._extract_sales_lines(order_sales, seats, item_value_discounts, discount_by_seat,
                                                   payment_by_seat, payment_by_benefit_key, used_benefit_keys,
                                                   discounts_tip_excluded, tip_excluded_by_seat)

            order = TableOrder(order_id, order_state, total_order_amount, total_order_after_discount,
                               total_order_tender, order_discount_amount, sale_lines, split_key, tip)
            table.orders.append(order)

        for benefit_key in payment_by_benefit_key.keys():
            self._add_payments_from_benefit_to_seat(
                    0,
                    benefit_key,
                    payment_by_seat,
                    payment_by_benefit_key,
                    used_benefit_keys
            )

        table.detailed_tip_excluded = tip_excluded_by_seat
        table.detailed_discounts = discount_by_seat
        table.detailed_payments = payment_by_seat
        table.sub_total_amount = table_order_total_gross + total_discount_amount
        table.tip = round_half_away_from_zero(total_tip, 2)
        total_amount = table_order_total_gross + table.tip - table.promotional_tenders
        table.total_amount = round_half_away_from_zero(total_amount, 2)

        return table

    @staticmethod
    def _get_table_name(table, table_xml):
        property_xml = table_xml.findall(".//Properties[@TAB_ID]")
        tab_name = property_xml[0].attrib["TAB_ID"] if property_xml else None
        table_name = table_xml.attrib["name"] if "name" in table_xml.attrib else table.table_id
        table.table_name = tab_name if tab_name else table_name

    @staticmethod
    def _get_table_attribute_value(table_xml_attrib, key, default_value):
        if key not in table_xml_attrib:
            return default_value

        return table_xml_attrib[key]

    def _get_discounts_from_sale_lines(self, sale_lines):
        # type: (List) -> Dict

        discount_values = {}
        for sale_line in sale_lines:
            discounts_applied = sale_line.get("discountsApplied") or None
            if discounts_applied is None:
                continue
            for discount in discounts_applied.split(','):
                discount_id, amount = discount.split(':')
                discount_id = int(discount_id)
                self._add_value_to_dict(discount_id, amount, discount_values)

        return discount_values

    def _get_discount_value(self, values_by_items, values_by_discount, tip_excluded_by_seat, discounts_tip_excluded,
                            sale_line_number, item_price, seat):
        # type: (Dict, Dict, Dict, List[int], int, float, int) -> float

        total_value = 0.0
        for discount_id in values_by_items.keys():
            for discount_line_number in values_by_items[discount_id]:
                if int(discount_line_number) == sale_line_number:
                    value = min(item_price, values_by_items[discount_id][discount_line_number])
                    values_by_items[discount_id][discount_line_number] -= value
                    values_by_discount[int(discount_id)] -= value
                    total_value += value

                    if int(discount_id) in discounts_tip_excluded:
                        self._add_value_to_dict(seat, value, tip_excluded_by_seat)

        return total_value

    @staticmethod
    def _get_service_period_abbreviation(properties):
        # type: (dict) -> Union[str, None]

        props = {}
        if 'servicePeriod' in properties:
            props = json.loads(properties['servicePeriod'])
        return props['servicePeriodAbbr'] if 'servicePeriodAbbr' in props else None

    def _get_parent_service_period(self, sales_line_list, line_number):
        # type: (List[eTree.Element], int) -> Union[str, None]

        for x in sales_line_list:
            if int(x.get("level")) != 0 or x.get("lineNumber") != line_number:
                continue
            props = x.get("customProperties")
            props = json.loads(props) if props else {}
            service_period = self._get_service_period_abbreviation(props)
            if service_period:
                return service_period
        return None

    def _get_service_period(self, sales_line_list, sale_line, properties):
        # type: (List[eTree.Element], eTree.Element, dict) -> str

        if int(sale_line.get("level")) == 0:
            return self._get_service_period_abbreviation(properties)
        else:
            item_service_period = self._get_parent_service_period(sales_line_list, sale_line.get("lineNumber"))

        return item_service_period

    def _extract_sales_lines(
            self,
            order_sales,                # type: List[eTree.Element]
            service_seats,              # type: int
            item_discount_values,       # type: Dict
            discount_by_seat,           # type: Dict
            payment_by_seat,            # type: Dict
            payment_by_benefit,         # type: Dict
            used_benefit_keys,          # type: List[str]
            discounts_tip_excluded,     # type: List[str]
            tip_excluded_by_seat        # type: Dict
    ):
        # type: (...) -> List[TableOrderSaleLine]

        deleted_line_numbers = map(lambda x: x.get("lineNumber"),
                                   filter(lambda x: x.get("level") == "0" and x.get("qty") == "0", order_sales))
        active_sale_lines = filter(lambda x: x.get("lineNumber") not in deleted_line_numbers, order_sales)

        values_by_discount = self._get_discounts_from_sale_lines(active_sale_lines)

        sales_line = []  # type: List[TableOrderSaleLine]

        for sale_line in active_sale_lines:
            products = sale_line.get("productName")
            price = float(sale_line.get("itemPrice") or 0)
            level = int(sale_line.get("level") or 0)
            qty = float(sale_line.get("qty") or 0)  if level == 0 else float(sale_line.get("multipliedQty") or 0)
            default_qty = int(sale_line.get("defaultQty")) if "defaultQty" in sale_line.attrib else 0
            item_type = sale_line.get("itemType")
            line_number = int(sale_line.get("lineNumber") or 0)
            comment = sale_line.find("Comment")
            if comment is not None:
                comment = comment.get("comment")

            seat = 0
            props = sale_line.get("customProperties")
            custom_properties = json.loads(props) if props else {}
            if 'seat' in custom_properties:
                seat = int(custom_properties['seat'])
                seat = 0 if seat > service_seats else seat

            service_period = self._get_service_period(active_sale_lines, sale_line, custom_properties)

            if 'benefitKey' in custom_properties:
                benefit_key = custom_properties['benefitKey']
                self._add_payments_from_benefit_to_seat(
                        seat,
                        benefit_key,
                        payment_by_seat,
                        payment_by_benefit,
                        used_benefit_keys
                )

            discount_value = self._get_discount_value(item_discount_values, values_by_discount, tip_excluded_by_seat,
                                                      discounts_tip_excluded, line_number, price, seat)
            self._add_value_to_dict(seat, discount_value, discount_by_seat)

            sale_line = TableOrderSaleLine(
                products, price, qty, default_qty, item_type, seat, level, comment, line_number, service_period
            )

            sales_line.append(sale_line)

        return sales_line

    def _get_promotional_tenders(self, benefit_list, table, payment_by_benefit_key, order_tenders):
        # type: (Dict, Table, Dict, List) -> None

        benefits = json.loads(benefit_list.get("benefitList"))
        benefits_with_tender = [x for x in benefits if x.get("promotion_tender")]
        for benefit in benefits_with_tender:
            benefit_key = benefit.get("benefit_key")
            for tender in benefit.get("promotion_tender"):
                amount = float(tender.get("amount"))
                self._add_value_to_dict(benefit_key, amount, payment_by_benefit_key)

        if not order_tenders or len(order_tenders) == 0:
            return

        tenders = {}
        for order_tender in order_tenders:
            tender_id = int(order_tender.get("tenderId", -1))
            amount = float(order_tender.get("tenderAmount", 0))
            self._add_value_to_dict(tender_id, amount, tenders)

        for benefit in benefits:
            for tender_id in benefit.get("order_tenders_ids"):
                amount = tenders.get(tender_id, 0)
                table.promotional_tenders += amount

    def _add_payments_from_benefit_to_seat(
            self,
            seat,                   # type: int
            benefit_key,            # type: str
            payment_by_seat,        # type: Dict
            payment_by_benefit,     # type: Dict
            used_benefit_keys       # type: List[str]
    ):
        # type: (...)-> None

        if benefit_key in used_benefit_keys or benefit_key not in payment_by_benefit:
            return

        self._add_value_to_dict(seat, payment_by_benefit[benefit_key], payment_by_seat)
        used_benefit_keys.append(benefit_key)

    @staticmethod
    def _add_value_to_dict(key, value, dict_by_key):
        # type: ([int, str], float, Dict) -> None

        if key in dict_by_key.keys():
            dict_by_key[key] += float(value)
        else:
            dict_by_key[key] = float(value)



from datetime import date

from actions.update import get_catalog_version_to_apply, get_message_update, apply_update
from actions.rupture.get_ruptured_products import has_rupture_items
from msgbus import TK_SYS_ACK
from sysactions import show_messagebox, format_date, StopAction

BUTTONS = "$OK|$CANCEL"


class OpenDayInteractor(object):
    @staticmethod
    def check_has_update_menu(pos_id):
        # type: (int) -> None

        today = date.today().strftime("%Y%m%d")
        confirm_message = "$CONFIRM_OPEN_CURRENT_DAY|{0}".format(format_date(today))
        msg = get_catalog_version_to_apply()
        if msg.token == TK_SYS_ACK:
            get_message_update(pos_id, "$APPLYING_UPDATE_MENU_VERSION_PENDING|" + msg.data)
            has_rupture_items(pos_id)
            applied_update = apply_update(pos_id)
            if not applied_update:
                confirm_message = "$UPDATE_MENU_ERROR_PROCEED"

        confirm = show_messagebox(pos_id, message=confirm_message, buttons=BUTTONS)
        if (confirm is None) or (confirm == 1):
            raise StopAction()



"""
For the ``future`` package.

Adds this import line:

    from __future__ import division

at the top and changes any old-style divisions to be calls to
past.utils.old_div so the code runs as before on Py2.6/2.7 and has the same
behaviour on Py3.

If "from __future__ import division" is already in effect, this fixer does
nothing.
"""

from lib2to3 import fixer_base
from lib2to3.fixer_util import syms, does_tree_import
from libfuturize.fixer_util import (token, future_import, touch_import_top,
                                    wrap_in_fn_call)


def match_division(node):
    u"""
    __future__.division redefines the meaning of a single slash for division,
    so we match that and only that.
    """
    slash = token.SLASH
    return node.type == slash and not node.next_sibling.type == slash and \
                                  not node.prev_sibling.type == slash


class FixDivisionSafe(fixer_base.BaseFix):
    # BM_compatible = True
    run_order = 4    # this seems to be ignored?

    _accept_type = token.SLASH

    PATTERN = """
    term<(not('/') any)+ '/' ((not('/') any))>
    """

    def start_tree(self, tree, name):
        """
        Skip this fixer if "__future__.division" is already imported.
        """
        super(FixDivisionSafe, self).start_tree(tree, name)
        self.skip = "division" in tree.future_features

    def match(self, node):
        u"""
        Since the tree needs to be fixed once and only once if and only if it
        matches, we can start discarding matches after the first.
        """
        if (node.type == self.syms.term and 
                    len(node.children) == 3 and
                    match_division(node.children[1])):
            expr1, expr2 = node.children[0], node.children[2]
            return expr1, expr2
        else:
            return False

    def transform(self, node, results):
        if self.skip:
            return
        future_import(u"division", node)

        touch_import_top(u'past.utils', u'old_div', node)
        expr1, expr2 = results[0].clone(), results[1].clone()
        # Strip any leading space for the first number:
        expr1.prefix = u''
        return wrap_in_fn_call("old_div", (expr1, expr2), prefix=node.prefix)




from abc import ABCMeta, abstractmethod
from ._DeviceServiceThread import DeviceServiceThread
from nperipherals_core.model.repository import NotificationRepository
from typing import Optional
from nperipherals_core.model.device_status import DeviceStatus


class Device(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_name(self):
        # type: () -> str
        raise NotImplementedError()

    @abstractmethod
    def initialize(self):
        # type: () -> None
        raise NotImplementedError()

    @abstractmethod
    def connect(self):
        # type: () -> None
        raise NotImplementedError()

    @abstractmethod
    def disconnect(self):
        # type: () -> None
        raise NotImplementedError()

    @abstractmethod
    def get_status(self):
        # type: () -> DeviceStatus
        raise NotImplementedError()

    def _create_service_thread(self, repository):
        # type: (NotificationRepository) -> Optional[DeviceServiceThread]
        pass

    def get_service_thread(self, repository):
        # type: (NotificationRepository) -> Optional[DeviceServiceThread]
        if getattr(self, "_service_thread", None) is None:
            # noinspection PyAttributeOutsideInit
            self._service_thread = self._create_service_thread(repository)
        return self._service_thread



import logging
import xml.sax

from cache_functions import UsersInfoCache
from xml.etree import cElementTree as eTree

from messageprocessorutil import UuidMessageProcessor
from mwrepository import ProductRepository
from production.manager import ProductionManager
from production.model import ProductionOrder

from ._SystemOrderParser import SystemOrderParser

performance_event_logger = logging.getLogger("ProductionPerformanceEvent")


class OrderModifiedProcessor(UuidMessageProcessor):
    def __init__(self, production_manager, product_repository, user_cache_thread=None):
        # type: (ProductionManager, ProductRepository, UsersInfoCache) -> None
        super(OrderModifiedProcessor, self).__init__()
        self.production_manager = production_manager
        self.part_code_jit_lines = product_repository.get_all_part_code_jit_lines()
        self.user_cache_thread = user_cache_thread

    def parse_data(self, data):
        self._performance_log(data)
        handler = SystemOrderParser(self.part_code_jit_lines, user_cache_thread=self.user_cache_thread)
        xml.sax.parseString(data, handler)
        if handler.order.order_id == 0:
            return None
        return handler.order

    @staticmethod
    def _performance_log(
        data,   # type: str
    ):
        # type (...) -> None

        try:
            data_xml = eTree.XML(data)
            order_id = (data_xml.find(".//Order") or {}).get("orderId", "")
            subject = data_xml.get("subject", "")
            event_type = data_xml.get("type", "")
            message = "Event - {:10} / {:20} / {}".format(order_id, subject, event_type)
            performance_event_logger.info(message)
        except Exception as _:
            performance_event_logger.exception("Error parsing event xml")

    def call_business(self, order):
        # type: (ProductionOrder) -> None
        if order is not None:
            self.production_manager.order_modified(order)

    def format_response(self, message_bus, message, event, input_obj, result):
        # no response
        pass

    def format_exception(self, message_bus, message, event, input_obj, exception):
        # no response
        pass

    def format_parse_exception(self, message_bus, message, event, data, exception):
        # no response
        pass



import socket
import sys
from abc import ABCMeta
from logging import Logger
from typing import Dict, Optional, List, Union

from nperipherals_core.model.custom_exceptions import DeviceNotConnected
from nperipherals_core.model.device import Device


class NetworkDevice(Device):
    __metaclass__ = ABCMeta

    def __init__(self, logger, device_configurations):
        # type: (Logger, Dict[unicode, unicode]) -> None

        self.logger = logger
        self.names = device_configurations["Name"]  # type: Union[List[str], str]
        self.host = device_configurations["Host"]  # type: unicode
        self.port = int(device_configurations["Port"])  # type: int

        self.socket = None  # type: Optional[socket.socket]
        self.SOCKET_BUF_SIZE = 1024

    def get_name(self):
        # type: () -> unicode
        if isinstance(self.names, list):
            return ','.join(self.names)
        return self.names

    def initialize(self):
        # type: () -> None

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._configure_socket()

    def connect(self):
        # type: () -> None

        try:
            self.socket.connect((self.host, self.port))
        except Exception as e:
            self.logger.exception("Error getting socket connection: [{}]".format(repr(e)))
            raise DeviceNotConnected()

    def disconnect(self):
        # type: () -> None

        try:
            self.socket.close()
        except Exception as _:
            self.logger.exception("Error closing socket connection")

    def _configure_socket(self):
        # type: () -> None

        keep_alive_on = 1
        idle_time_before_keep_alive_seconds = 1
        interval_between_keep_alive_packets_seconds = 3
        keep_alive_max_retries = 5

        if sys.platform == "win32":
            self._configure_windows_socket_keepalive(idle_time_before_keep_alive_seconds,
                                                     interval_between_keep_alive_packets_seconds,
                                                     keep_alive_on)
        else:
            self._configure_linux_socket_keepalive(idle_time_before_keep_alive_seconds,
                                                   interval_between_keep_alive_packets_seconds,
                                                   keep_alive_max_retries,
                                                   keep_alive_on)

    def _configure_linux_socket_keepalive(self, idle_time_before_keep_alive_seconds,
                                          interval_between_keep_alive_packets_seconds,
                                          keep_alive_max_retries,
                                          keep_alive_on):
        # type: (int, int, int, int) -> None

        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, keep_alive_on)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, idle_time_before_keep_alive_seconds)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_between_keep_alive_packets_seconds)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, keep_alive_max_retries)

    def _configure_windows_socket_keepalive(self, idle_time_before_keep_alive_seconds,
                                            interval_between_keep_alive_packets_seconds,
                                            keep_alive_on):
        # type: (int, int, int) -> None

        self.socket.ioctl(socket.SIO_KEEPALIVE_VALS, (keep_alive_on,
                                                      1000 * idle_time_before_keep_alive_seconds,
                                                      1000 * interval_between_keep_alive_packets_seconds))



# -*- coding: utf-8 -*-

from json import JSONEncoder

from application.model.benefit import BenefitList
from typing import Dict, Any


class NDiscountModelJsonEncoder(JSONEncoder):
    def default(self, o):
        instances_dictionary = {
            BenefitList: self._get_benefit_list_json
        }

        if type(o) in instances_dictionary:
            return instances_dictionary[type(o)](o)

        return super(NDiscountModelJsonEncoder, self).default(o)

    @staticmethod
    def _get_benefit_list_json(o):
        # type: (BenefitList) -> Dict[str, Any]

        return {
            "benefitPreTotalSequence": o.benefit_pre_total_sequence,
            "manualBenefitOnTotalSequence": o.manual_benefits_on_total_sequence,
            "discountItemValues": o.discount_item_values,
            "discountIdTipExcluded": o.discount_id_tip_excluded,
            "tenderIdTipExcluded": o.tender_id_tip_excluded,
            "benefitList": o.benefit_list
        }



from commons.report import Generator
from deliveryreport import DeliveryReport
from dto import DeliveryHeaderDto, DeliveryBodyDto, DeliveryGeneratorDto


class DeliveryGenerator(Generator):
    def __init__(self, order_repository, product_repository, order_id):
        self.order_repository = order_repository
        self.product_repository = product_repository
        self.order_id = order_id

    def generate_data(self):
        delivery_report = DeliveryReport(self.order_repository, self.product_repository)
        order = delivery_report.get_delivery_order(self.order_id)
        products = delivery_report.get_products()
        cfh_products = delivery_report.get_cfh_products()
        header = DeliveryHeaderDto(order)
        body = DeliveryBodyDto(order)
        return DeliveryGeneratorDto(header, body, products, cfh_products)



from abc import ABCMeta, abstractmethod


class TableService(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_table_picture(self, pos_id, table_id):
        # type: (str, str) -> str
        raise NotImplementedError()

    @abstractmethod
    def get_tip_rate(self):
        # type: () -> int
        raise NotImplementedError()



# -*- coding: utf-8 -*-
import logging
import json
from datetime import datetime, timedelta

from bustoken import TK_SITEF_ADD_PAYMENT, TK_SITEF_FINISH_PAYMENT, TK_SITEF_AVAILABLE
from messagehandler import EventHandler
from msgbus import TK_SYS_NAK, TK_SYS_ACK
from helper import remove_accents

logger = logging.getLogger("Cappta")


class CapptaEventHandler(EventHandler):
    def __init__(self, mbcontext, cappta_processor, electronic_payments):
        super(CapptaEventHandler, self).__init__(mbcontext)
        self.cappta_processor = cappta_processor

        self.electronic_payments = electronic_payments

    def get_handled_tokens(self):
        return [TK_SITEF_ADD_PAYMENT, TK_SITEF_FINISH_PAYMENT, TK_SITEF_AVAILABLE]

    def handle_message(self, msg):
        try:
            if msg.token == TK_SITEF_ADD_PAYMENT:
                self._handle_add_payment(msg)
            elif msg.token == TK_SITEF_FINISH_PAYMENT:
                self._handle_finish_payment(msg)
            elif msg.token == TK_SITEF_AVAILABLE:
                self._handle_sitef_available(msg)
            else:
                msg.token = TK_SYS_NAK
                self.mbcontext.MB_ReplyMessage(msg)

        except Exception as _:
            logger.exception("Unexpected exception when handling event {0}".format(msg.token))
            msg.token = TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg)

    def handle_event(self, subject, evt_type, data, msg):
        raise NotImplementedError()

    def terminate_event(self):
        return

    def _handle_sitef_available(self, msg):
        msg.token = TK_SYS_ACK
        msg.data = json.dumps(self.electronic_payments)
        self.mbcontext.MB_ReplyMessage(msg, data=msg.data)

    def _handle_add_payment(self, msg):
        posid, orderid, operador, tender_type, amount, data_fiscal, hora_fiscal, tender_seq_id, display_via_api = msg.data.split(';')
        result, msg.data = self.cappta_processor.process(msg, posid, orderid, operador, tender_type, amount, data_fiscal, hora_fiscal, tender_seq_id)
        msg.token = TK_SYS_ACK
        self.mbcontext.MB_ReplyMessage(msg, data=msg.data)

    def _handle_cancel_payment(self, msg):
        msg.data = self.cappta_processor.cancel()
        msg.token = TK_SYS_ACK
        self.mbcontext.MB_ReplyMessage(msg, data=msg.data)

    def _handle_finish_payment(self, msg):
        data = msg.data.split(';')
        posid, orderid, data_fiscal, hora_fiscal, status = data[0:5]
        msg.data = self.cappta_processor.finalizar(posid, orderid, data_fiscal, hora_fiscal, int(status))
        msg.token = TK_SYS_ACK
        self.mbcontext.MB_ReplyMessage(msg, data=msg.data)


class CapptaCallback(object):
    def __init__(self, mbcontext):
        self.mbcontext = mbcontext
        self.last_response = dict(message=None, timestamp=None)

    def callback(self, bus_msg, pos_id, response, message):
        from sysactions import show_messagebox, show_keyboard

        try:
            formatted_response = self._format_response(response)

            is_repeated = self._check_if_message_is_not_repeated(response)
            if is_repeated:
                return

            logger.info("Callback -> PosID: {}; Type: {}; Message: {}".format(pos_id, message, formatted_response))

            if message == "MENSAGEM":
                data = ";".join([formatted_response, pos_id])
                self.mbcontext.MB_EasyEvtSend("SITEF_STATUS_UPDATE", "STATUS_UPDATE", data)

            elif message in ["PAGAMENTO_CONFIRMADO", "OPERACAO_CANCELADA", "FALHA_AUTENTICACAO"]:
                msg = bus_msg
                msg.token = 242424
                msg.data = formatted_response
                self.mbcontext.MB_ReplyMessage(msg, data=msg.data)

            elif message == "REQUISICAO_NUMERO_CARTAO":
                param = "CANCELA"
                return str(param)

            elif message in ["REQUISICAO_VALIDADE", "REQUISICAO_CODIGO_SEGURANÇA"]:
                formatted_response = formatted_response.split("\n")
                param = show_keyboard(pos_id, formatted_response[1], title=formatted_response[0], mask="INTEGER", numpad=True, buttons="$OK|$CANCEL")
                param = "CANCELA" if param is None else param
                return str(param)

            elif message == "REQUISICAO_ACAO":
                message_box = "Cancelar operação?"
                formatted_response = formatted_response.split("\n")
                param = show_messagebox(pos_id, message_box, title=formatted_response[0], buttons="$YES|$NO")
                if param == 0:
                    param = 1
                else:
                    param = 0
                return str(param)

            elif message == "REQUISICAO_DESCONECTAR_PINPAD":
                message_box = "Pinpad não encontrado. Por favor reconecte-o.\nTentar novamente?"
                param = show_messagebox(pos_id, message_box, title="", buttons="$YES|$NO")
                if param == 0:
                    param = 1
                else:
                    param = 0
                return str(param)
            elif message == "TRANSACAO_PENDENTE":
                message_box = "Última transação ainda pendente.\nDeseja confirmá-la?"
                param = show_messagebox(pos_id, message_box, title="", buttons="Confirmar|$CANCEL")
                if param == 0:
                    param = 1
                else:
                    param = 0
                return str(param)
        except BaseException as _:
            logger.exception("Error on callback")

    def _check_if_message_is_not_repeated(self, response):
        now = datetime.now()
        if response == self.last_response["message"] and (now - self.last_response["timestamp"]) < timedelta(seconds=3):
            return True
        else:
            self.last_response["message"] = response
            self.last_response["timestamp"] = now
            return False

    @staticmethod
    def _format_response(response):
        return remove_accents(response)



from logging import Logger
from threading import Thread

from _ProductionView import ProductionView
from messagebus import MessageBus, Event
from production.model import ProductionOrder
from production.view._OrderXml import OrderXml
from typing import List


class PublishView(ProductionView):
    def __init__(self, name, message_bus, event, event_type, tag, send_order, logger):
        # type: (str, MessageBus, str, str, str, bool, Logger) -> None

        super(PublishView, self).__init__(name, logger)
        self.message_bus = message_bus
        self.event = event
        self.event_type = event_type
        self.tag = tag
        self.logger = logger
        self.send_order = send_order
        self.order_xml = OrderXml()

    def handle_order(self, order):
        # type: (ProductionOrder) -> Thread

        t = Thread(target=self.publish_event, args=[order])
        t.daemon = True
        t.start()
        return t

    def get_view_tags(self):
        # type: () -> List[str]

        return [self.tag]

    def publish_event(self, order):
        # type: (ProductionOrder) -> None

        try:
            if self.event_type == "SingleEvent" and self.some_item_was_tagged(order.items):
                return

            if self.send_order:
                event = Event(self.event, "", self.order_xml.to_xml(order))
            else:
                event = Event(self.event, "", str(order.order_id).encode("utf-8"))

            self.logger.info("Publishing event: {}".format(event.subject))
            self.message_bus.publish_event(event)
        except BaseException as _:  # noqa
            self.exception("Exception when publishing event")

    def some_item_was_tagged(self, items):
        for item in items:
            if self.tag and self.tag in item.get_tags():
                return True

            if len(item.items) > 0 and self.some_item_was_tagged(item.items):
                return True

        return False



from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder
from production.box import FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBox


class FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBox
        return FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBox(config.name,
                                                                    config.sons,
                                                                    self.get_extra(config, "AllowedTags", []),
                                                                    self.get_extra(config, "ExcludedTags", []),
                                                                    self.get_extra(config, "ForbiddenTags", []),
                                                                    self.get_extra(config, "AllowNoTags", "True"),
                                                                    self.get_extra(config, "FilterType", "byOrder"),
                                                                    self.get_logger(config))



from old_helper import BaseRepository
from datetime import datetime
from persistence import Connection
from ..model import Transfer
from typing import Optional, List


class AccountRepository(BaseRepository):
    TypeRealDate = "RealDate"
    TypeBusinessPeriod = "BusinessPeriod"
    TypeSessionId = "SessionId"

    def __init__(self, mb_context):
        super(AccountRepository, self).__init__(mb_context)

    def get_transfers_by_real_date(self, initial_date, end_date, operator_id, report_pos, session_id=None):
        # type: (datetime, datetime, str, str, Optional[unicode]) -> List[Transfer]
        return self._get_transfer(self.TypeRealDate, initial_date, end_date, operator_id, session_id, report_pos)

    def get_transfers_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, str, str) -> List[Transfer]
        return self._get_transfer(self.TypeBusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_transfers_by_session_id(self, session_id):
        # type: (unicode) -> List[Transfer]
        return self._get_transfer(self.TypeSessionId, None, None, None, session_id, None)

    def _get_transfer(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode], Optional[unicode]) -> List[Transfer] # noqa
        def inner_func(conn):
            # type: (Connection) -> List[Transfer]
            current_session_id = session_id if session_id is not None else ''
            if report_type == self.TypeRealDate or report_type == self.TypeBusinessPeriod:
                query = self._TransfersByRealDate
                if report_type == self.TypeBusinessPeriod:
                    query = self._TransfersByBusinessPeriod

                formatted_initial_date = initial_date.strftime("%Y-%m-%d")
                formatted_end_date = end_date.strftime("%Y-%m-%d")
                query = query.format(formatted_initial_date, formatted_end_date, current_session_id)

                if operator_id is not None:
                    query += " and SessionId like '%user={0},%'".format(operator_id)

                if report_pos is not None:
                    query += " and PosId = {0}".format(report_pos)
            else:
                query = self._TransfersBySessionId
                query = query.format(session_id)


            return [Transfer(int(x.get_entry(0)), float(x.get_entry(1))) for x in conn.select(query)]

        return self.execute_with_connection(inner_func)

    _TransfersByRealDate = """
                                SELECT Type, Amount 
                                FROM Transfer
                                WHERE strftime('%Y-%m-%d', Timestamp, 'localtime') >= '{0}' AND
                                strftime('%Y-%m-%d', Timestamp, 'localtime') <= '{1}' AND
                                SessionId LIKE '%{2}%'
                            """

    _TransfersByBusinessPeriod = """
                                    SELECT Type, Amount 
                                    FROM Transfer 
                                    WHERE Period >= strftime('%Y%m%d', '{0}') AND 
                                          Period <= strftime('%Y%m%d', '{1}') AND SessionId LIKE '%{2}%'
                                """

    _TransfersBySessionId = """
                                SELECT Type, Amount 
                                FROM Transfer
                                WHERE SessionId = '{0}'
                            """



from production.box import HideOptionBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class HideOptionBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(HideOptionBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> HideOptionBox
        return HideOptionBox(config.name,
                             config.sons,
                             self.get_extra(config, "FixDefaultQuantity", "false").lower() == "true",
                             self.get_logger(config))



from typing import Optional  # noqa
import datetime  # noqa


class PickListHeaderDto(object):
    def __init__(self, date, order_id, store, client_name):
        # type: (datetime, int, unicode, Optional[unicode]) -> ()
        self.date = date
        self.order_id = order_id
        self.store = store
        self.client_name = client_name



from typing import Any  # noqa
import qrcode


class QrCodeGenerator(object):
    @staticmethod
    def generate_qr_code(content):
        # type: (unicode) -> Any
        qr_code = qrcode.QRCode(version=2,
                                box_size=6,
                                border=1,
                                error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr_code.add_data(content)
        qr_code.make(fit=True)

        return qr_code.make_image().convert("RGB")



from sysactions import get_user_information
from xml.etree import cElementTree as eTree

from model.customexception import NotFoundUserError


def get_user_level(user_id):
    if user_id is None:
        return None

    user_xml_str = get_user_information(user_id)
    if user_xml_str is None:
        raise NotFoundUserError("USER_NOT_FOUND")
    
    user_xml = eTree.XML(user_xml_str)
    user_element = user_xml.find("user")
    user_level = int(user_element.attrib["Level"])
    
    return user_level



# encoding: utf-8
import sys

from typing import Dict, List
from threading import Thread, Condition
from msgbus import \
    MBException, \
    SE_SUCCESS, \
    TK_CMP_TERM_NOW, \
    TK_CMP_START, \
    TK_EVT_EVENT, \
    TK_SYS_NAK, \
    TK_SYS_ACK

from syserrors import SE_MB_ERR
from systools import sys_log_exception, sys_log_error

from messagebus import Event, Message, DataType
from messagehandler import MessageHandlerBuilder
from _MbContextMessageBus import MbContextMessageBus


class MbContextMessageHandler(object):
    def __init__(self, message_bus, service_name, service_type, required_services, event_handler_builder):
        # type: (MbContextMessageBus, str, str, str, MessageHandlerBuilder) -> None
        self.message_bus = message_bus
        self.service_name = service_name
        self.service_type = service_type
        self.required_services = required_services
        self.event_handler_builder = event_handler_builder
        self.reentrant_subjects = {}
        self.non_reentrant_subjects = {}
        self.queue_subjects = {}
        self.sync_subjects = {}
        self.thread_handling_subject = {}
        self.subject_event_queue = {}  # type: Dict[unicode, List[Event]]
        self.subject_queue_thread = {}  # type: Dict[unicode, Thread]
        self.subject_queue_condition = {}  # type: Dict[unicode, Condition]

        self.tokens = {}
        self.finished = False

        try:
            # Se temos um servico, registramos o mesmo
            if self.service_name and self.service_type != "":
                status = self.message_bus.mbcontext.MB_EasyWaitStart("{}:{}"
                                                                     .format(self.service_name, self.service_type),
                                                                     self.required_services)
                if status != SE_SUCCESS:
                    sys_log_error("Terminating component after 'MB_EasyWaitStart'")
                    sys.exit(status)
        except MBException:
            sys_log_exception("Message-bus error while initializing the FiscalWrapper component. Terminating")
            sys.exit(SE_MB_ERR)

    def subscribe_reentrant_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler
        for subject in subjects:
            self.reentrant_subjects[subject] = subject
            self.message_bus.subscribe(subject)

        return self

    def subscribe_non_reentrant_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler
        for subject in subjects:
            self.non_reentrant_subjects[subject] = subject
            self.message_bus.subscribe(subject)

        return self

    def subscribe_queue_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler
        for subject in subjects:
            self.queue_subjects[subject] = subject

            self.subject_event_queue[subject] = []
            self.subject_queue_condition[subject] = Condition()
            subject_queue_thread = Thread(target=self.event_queue_method,
                                          name=str(subject) + " - Event Queue Thread",
                                          args=(subject,))
            subject_queue_thread.daemon = True
            subject_queue_thread.start()
            self.subject_queue_thread[subject] = subject_queue_thread

            self.message_bus.subscribe(subject)

        return self

    def subscribe_sync_events(self, subjects):
        # type: (List[unicode]) -> MbContextMessageHandler
        for subject in subjects:
            self.sync_subjects[subject] = subject
            self.message_bus.subscribe(subject)

        return self

    def handle_events(self):
        self.event_handler_builder.build_singletons()

        while not self.finished:
            msg = self.message_bus.mbcontext.MB_EasyGetMessage()

            # Check for component terminate request
            if (not msg) or (msg.token == TK_CMP_TERM_NOW):
                # Component terminated, warn the handler that we are terminated
                self.event_handler_builder.destroy_singletons()
                # reply the message
                self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                # Finish the loop and terminate the component
                self.finished = True
                break

            if msg.token == TK_CMP_START:
                msg.token = TK_SYS_ACK
                self.message_bus.mbcontext.MB_EasyReplyMessage(msg)

            elif msg.token == TK_EVT_EVENT:
                params = msg.data.split('\x00')
                data = params[0] if len(params) > 0 else None
                subject = params[1] if len(params) > 1 else None
                evt_type = params[2] if len(params) > 2 else None
                if evt_type == '':
                    evt_type = None
                sync = params[3] == 'true' if len(params) > 3 else False
                source_id = int(params[4]) if len(params) > 4 else None

                event = Event(subject, evt_type, data, source_id, sync)
                event.imp_message = msg

                if subject in self.non_reentrant_subjects and \
                        subject in self.thread_handling_subject and \
                        self.thread_handling_subject[subject].is_alive():
                    # Se o evento é não reentrante e já temos uma thread tratando outra instância do mesmo evento,
                    # respondemos e não criamos outra thread
                    self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                    continue

                if subject in self.queue_subjects:
                    # Se o evento é do tipo fila, enfileiramos a tupla e deixamos a thread responsável tratá-lo
                    with self.subject_queue_condition[subject]:
                        self.subject_event_queue[subject].append(event)
                        self.subject_queue_condition[subject].notify()

                    self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
                    continue

                handler_thread = Thread(target=self._handle_event, args=(event, ))
                handler_thread.daemon = True

                if subject in self.non_reentrant_subjects:
                    self.thread_handling_subject[subject] = handler_thread

                handler_thread.start()

                # Se o evento for syncrono, deixamos a thread responsável por dar o reply na mensagem
                if subject not in self.sync_subjects:
                    self.message_bus.mbcontext.MB_EasyReplyMessage(msg)
            # Check for a event request
            else:
                # handles the event ...
                data = msg.data
                data_format = msg.format
                if msg.data is None or msg.data == "":
                    data = None
                    data_format = DataType.empty

                message = Message(msg.token, data_format, data)
                message.imp_message = msg
                handler_thread = Thread(target=self._handle_message, args=(message,))
                handler_thread.daemon = True
                handler_thread.start()

    def event_queue_method(self, subject):
        my_condition = self.subject_queue_condition[subject]
        my_queue = self.subject_event_queue[subject]

        while not self.finished:
            while len(my_queue) > 0:
                with my_condition:
                    event = my_queue.pop(0)

                try:
                    self._handle_event(event)
                except:
                    pass

            with my_condition:
                my_condition.wait()

    def _handle_event(self, event):
        # type: (Event) -> None
        event_handler = None
        try:
            event_handler = self.event_handler_builder.build_message_handler()
            event_handler.handle_event(self.message_bus, event)
        finally:
            if event_handler is not None:
                self.event_handler_builder.destroy_message_handler(event_handler)

    def _handle_message(self, message):
        # type: (Message) -> None
        event_handler = None
        try:
            event_handler = self.event_handler_builder.build_message_handler()
            event_handler.handle_message(self.message_bus, message)
        finally:
            if event_handler is not None:
                self.event_handler_builder.destroy_message_handler(event_handler)




class OperatorTip(object):
    def __init__(self, operator, service_id, default_unit_price, tip_portion):
        # type: (str, int, float, float) -> None
        self.operator = operator
        self.service_id = service_id
        self.default_unit_price = round(default_unit_price, 2)
        self.tip_portion = tip_portion



MENU_MANAGAR_AUTHENTICATED_USER = None

def get_menu_manager_authenticated_user():
    return MENU_MANAGAR_AUTHENTICATED_USER

def set_menu_manager_authenticated_user(user_id):
    global MENU_MANAGAR_AUTHENTICATED_USER
    MENU_MANAGAR_AUTHENTICATED_USER = user_id




# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""Module: barcode.upc

:Provided barcodes: UPC-A
"""
__docformat__ = 'restructuredtext en'

from barcode.base import Barcode
from barcode.charsets import upc as _upc
from barcode.errors import *

class UniversalProductCodeA(Barcode):
    """Initializes new UPC-A barcode.

    :parameters:
        upc : String
            The upc number as string.
        writer : barcode.writer Instance
            The writer to render the barcode (default: SVGWriter).
        make_ean: boolean
    """

    name = 'UPC-A'

    digits = 11

    def __init__(self, upc, writer=None, make_ean=False):
        self.ean = make_ean
        upc = upc[:self.digits]
        if not upc.isdigit():
            raise IllegalCharacterError('UPC code can only contain numbers.')
        if len(upc) != self.digits:
            raise NumberOfDigitsError('UPC must have {0} digits, not '
                                      '{1}.'.format(self.digits, len(upc)))
        self.upc = upc
        self.upc = '{}{}'.format(upc, self.calculate_checksum())
        self.writer = writer or Barcode.default_writer()

    def __unicode__(self):
        if self.ean:
            return '0' + self.upc
        else:
            return self.upc

    __str__ = __unicode__

    def get_fullcode(self):
        if self.ean:
            return '0' + self.upc
        else:
            return  self.upc

    def calculate_checksum(self):
        """Calculates the checksum for UPCA/UPC codes

        :return: The checksum for 'self.upc'
        :rtype: Integer
        """
        def sum_(x, y): return int(x) + int(y)
        upc = self.upc[0:self.digits]
        oddsum = reduce(sum_, upc[::2])
        evensum = reduce(sum_, upc[1::2])
        check = (evensum + oddsum * 3) % 10
        if check == 0:
            return 0
        else:
            return 10 - check


    def build(self):
        """Builds the barcode pattern from 'self.upc'

        :return: The pattern as string
        :rtype: String
        """
        code = _upc.EDGE[:]

        for i, number in enumerate(self.upc[0:6]):
            code += _upc.CODES['L'][int(number)]

        code += _upc.MIDDLE

        for number in self.upc[6:]:
            code += _upc.CODES['R'][int(number)]

        code += _upc.EDGE

        return [code]

    def to_ascii(self):
        """Returns an ascii representation of the barcode.

        :rtype: String
        """

        code = self.build()
        for i, line in enumerate(code):
            code[i] = line.replace('1', '|').replace('0', '_')
        return '\n'.join(code)

    def render(self, writer_options=None):
        options = dict(module_width=0.33)
        options.update(writer_options or {})
        return Barcode.render(self, options)

UPCA = UniversalProductCodeA



import decimal
from decimal import Decimal
from unittest import TestCase
from simplejson.compat import StringIO, reload_module

import simplejson as json

class TestDecimal(TestCase):
    NUMS = "1.0", "10.00", "1.1", "1234567890.1234567890", "500"
    def dumps(self, obj, **kw):
        sio = StringIO()
        json.dump(obj, sio, **kw)
        res = json.dumps(obj, **kw)
        self.assertEqual(res, sio.getvalue())
        return res

    def loads(self, s, **kw):
        sio = StringIO(s)
        res = json.loads(s, **kw)
        self.assertEqual(res, json.load(sio, **kw))
        return res

    def test_decimal_encode(self):
        for d in map(Decimal, self.NUMS):
            self.assertEqual(self.dumps(d, use_decimal=True), str(d))

    def test_decimal_decode(self):
        for s in self.NUMS:
            self.assertEqual(self.loads(s, parse_float=Decimal), Decimal(s))

    def test_stringify_key(self):
        for d in map(Decimal, self.NUMS):
            v = {d: d}
            self.assertEqual(
                self.loads(
                    self.dumps(v, use_decimal=True), parse_float=Decimal),
                {str(d): d})

    def test_decimal_roundtrip(self):
        for d in map(Decimal, self.NUMS):
            # The type might not be the same (int and Decimal) but they
            # should still compare equal.
            for v in [d, [d], {'': d}]:
                self.assertEqual(
                    self.loads(
                        self.dumps(v, use_decimal=True), parse_float=Decimal),
                    v)

    def test_decimal_defaults(self):
        d = Decimal('1.1')
        # use_decimal=True is the default
        self.assertRaises(TypeError, json.dumps, d, use_decimal=False)
        self.assertEqual('1.1', json.dumps(d))
        self.assertEqual('1.1', json.dumps(d, use_decimal=True))
        self.assertRaises(TypeError, json.dump, d, StringIO(),
                          use_decimal=False)
        sio = StringIO()
        json.dump(d, sio)
        self.assertEqual('1.1', sio.getvalue())
        sio = StringIO()
        json.dump(d, sio, use_decimal=True)
        self.assertEqual('1.1', sio.getvalue())

    def test_decimal_reload(self):
        # Simulate a subinterpreter that reloads the Python modules but not
        # the C code https://github.com/simplejson/simplejson/issues/34
        global Decimal
        Decimal = reload_module(decimal).Decimal
        import simplejson.encoder
        simplejson.encoder.Decimal = Decimal
        self.test_decimal_roundtrip()



# encoding: utf-8
import os
import six

from typing import Optional
from PIL import Image, ImageOps
from commons.util import EscPosImage
from constants import ESC, GS


class LogoImage(object):
    def __init__(self):
        # type: () -> None
        self.data_dir = os.environ["HVDATADIR"]

    def get_logo(self):
        # type: () -> Image

        image_logo = self._logo_path()
        if image_logo is None:
            return ""

        base_width = 250
        with Image.open(image_logo) as image_opened:
            width_percent = (base_width / float(image_opened.size[0]))
            height_size = int((float(image_opened.size[1]) * float(width_percent)))
            with image_opened.resize((base_width, height_size)) as image_resize:
                with image_resize.convert("RGBA") as image_converted:
                    return self._image(image_converted, impl="bitImageColumn")

    def _logo_path(self):
        # type: () -> Optional[str]

        images_path = os.path.join(self.data_dir, "../../src/edpreports/src/static/logo_footer.png")
        if not os.path.exists(images_path):
            return None

        return images_path

    def _image(self, img_source, high_density_vertical=True, high_density_horizontal=True, impl="bitImageRaster"):
        im = EscPosImage(img_source)
        image_print = ""
        if impl == "bitImageRaster":
            # GS v 0, raster format bit image
            density_byte = (0 if high_density_horizontal else 1) + (0 if high_density_vertical else 2)
            header = GS + b"v0" + six.int2byte(density_byte) + self._int_low_high(im.width_bytes, 2) + self._int_low_high(
                im.height,
                2)
            image_print += (header + im.to_raster_format())
            return image_print

        if impl == "bitImageColumn":
            # ESC *, column format bit image
            density_byte = (1 if high_density_horizontal else 0) + (32 if high_density_vertical else 0)
            header = ESC + b"*" + six.int2byte(density_byte) + self._int_low_high(im.width, 2)
            outp = [ESC + b"3" + six.int2byte(48)]  # Adjust line-feed size
            for blob in im.to_column_format(high_density_vertical):
                outp.append(header + blob + b"\n")
            outp.append(ESC + b"2")  # Reset line-feed size
            return b''.join(outp)

    @staticmethod
    def _int_low_high(inp_number, out_bytes):
        """ Generate multiple bytes for a number: In lower and higher parts, or more parts as needed.

        :param inp_number: Input number
        :param out_bytes: The number of bytes to output (1 - 4).
        """
        max_input = (256 << (out_bytes * 8) - 1)
        if not 1 <= out_bytes <= 4:
            raise ValueError("Can only output 1-4 byes")
        if not 0 <= inp_number <= max_input:
            raise ValueError("Number too large. Can only output up to {0} in {1} byes".format(max_input, out_bytes))
        outp = b''
        for _ in range(0, out_bytes):
            outp += six.int2byte(inp_number % 256)
            inp_number //= 256
        return outp



import json
from xml.etree import cElementTree as eTree

from application.model import DefaultBenefitApplierRepository, LoyaltyTokens
from application.model.customexception import ErrorBurningCoupon, ErrorUnlockingCoupon
from application.model.customexception._CustomException import ErrorCheckingVoucher
from messagebus import Message, DataType
from msgbus import TK_SYS_ACK, TK_SYS_NAK


class LoyaltyBenefitApplierRepository(DefaultBenefitApplierRepository):

    def __init__(self, configs, message_bus):
        super(LoyaltyBenefitApplierRepository, self).__init__(configs, message_bus)

        self.benefit = {}

    def check_voucher(self, pos_id, voucher_id):
        # type: (int, str) -> bool

        message = Message(
            token=LoyaltyTokens.TK_LOYALTY_GET_AND_LOCK_VOUCHER,
            data=json.dumps(dict(posId=pos_id, voucherId=voucher_id)),
            data_type=DataType.string
        )

        response = self.message_bus.send_message("Loyalty", message)
        valid_response_data = response.data is not None and len(response.data) > 0 and response.data.lower() != "none"
        if response.token == TK_SYS_ACK and valid_response_data and "error" not in response.data.lower():
            self.benefit[voucher_id] = response.data
            return True

        if response.data:
            if "error:" in response.data.lower():
                raise ErrorCheckingVoucher(response.data.split(":")[-1])

            raise ErrorCheckingVoucher(response.data)

        return False

    def get_benefit_and_lock(self, voucher_id):
        # type: (str) -> None

        benefit = self.benefit[voucher_id]
        del self.benefit[voucher_id]

        return benefit

    def unlock_benefit(self, benefit_id):
        # type: (str) -> None

        message = Message(
            token=LoyaltyTokens.TK_LOYALTY_UNLOCK_VOUCHER,
            data=benefit_id,
            data_type=DataType.string
        )

        response = self.message_bus.send_message("Loyalty", message)
        if response.token == TK_SYS_NAK:
            raise ErrorUnlockingCoupon(response.data)

    def burn_benefit(self, benefit_id):
        # type: (str) -> None

        message = Message(
            token=LoyaltyTokens.TK_LOYALTY_BURN_VOUCHER,
            data=benefit_id,
            data_type=DataType.string
        )

        response = self.message_bus.send_message("Loyalty", message)
        if response.token == TK_SYS_NAK:
            raise ErrorBurningCoupon(response.data)

    def get_benefit_info(self, data):
        # type: (str) -> None

        return

    def get_benefit_id_by_custom_property(self, order_picture):
        # type: (eTree) -> str

        order_properties = order_picture.findall(".//OrderProperty")

        for prop in order_properties:
            key = prop.get("key")

            if key == "LOYALTY_ID":
                return prop.get("value")



from mw_helper import convert_to_dict, ensure_list
from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class PodTypeFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, allowed_pod_types, excluded_pod_types=None, logger=None):
        super(PodTypeFilterBox, self).__init__(name, sons, logger)
        if allowed_pod_types is None:
            allowed_pod_types = []
        self.allowed_pod_types = convert_to_dict(ensure_list(allowed_pod_types))
        if excluded_pod_types is None:
            excluded_pod_types = []
        self.excluded_pod_types = convert_to_dict(ensure_list(excluded_pod_types))

    def change_order(self, order):
        if self.excluded_pod_types:
            if order.pod_type in self.excluded_pod_types:
                return None

            return order

        if order.pod_type not in self.allowed_pod_types:
            return None

        return order



# -*- coding: utf-8 -*-

import sysactions
from actions.rupture.get_ruptured_products import has_rupture_items

from bustoken import TK_POS_UPDATER_PERFORM_CATALOG_UPDATE, TK_POS_GET_CATALOG_VERSION_TO_APPLY
from msgbus import TK_SYS_NAK, FM_PARAM


@sysactions.action
def update_menu(pos_id):
    try:
        msg = get_catalog_version_to_apply(True)
        if msg.token == TK_SYS_NAK:
            get_message_update(pos_id, "$UPDATE_MENU_" + msg.data.upper(), buttons="OK")
            raise sysactions.StopAction()

        get_message_update(pos_id, "$UPDATE_MENU_VERSION_AVAILABLE|" + msg.data, buttons="$OK|$CANCEL")
        get_message_update(pos_id, "$UPDATE_MENU_MESSAGE", buttons="$OK|$CANCEL")
        has_rupture_items(pos_id)
        apply_update(pos_id)
    except Exception as _:
        sysactions.sys_log_exception("Error updating menu")
        get_message_update(pos_id, "$UPDATE_MENU_EXCEPTION", buttons="OK")


def get_catalog_version_to_apply(need_backoffice_json=False):
    timeout = 60000000 * 5
    data = str(need_backoffice_json)
    return sysactions.send_message("PosUpdater", TK_POS_GET_CATALOG_VERSION_TO_APPLY, FM_PARAM, data, timeout)


def apply_update(pos_id):
    timeout = 60000000 * 5
    msg = sysactions.send_message("PosUpdater", TK_POS_UPDATER_PERFORM_CATALOG_UPDATE, FM_PARAM, timeout=timeout)
    if msg.token == TK_SYS_NAK:
        get_message_update(pos_id, "$UPDATE_MENU_" + msg.data.upper(), buttons="OK")
        return False

    return True


def get_message_update(pos_id, message, buttons="OK", timeout=60000):
    resp = sysactions.show_messagebox(pos_id, message, buttons=buttons, timeout=timeout)
    if resp is None or resp == 1:
        raise sysactions.StopAction()



# -*- coding: utf-8 -*-
import logging

from sysactions import StopAction, show_messagebox, set_custom, get_business_period, get_operator_session
from actions.util import get_drawer_amount

logger = logging.getLogger("PosActions")


def do_set_drawer_status(pos_id, amount, state=None, alert_level_1='0'):
    from posactions import sangria_levels

    try:
        sangria_levels_list = [float(level) for level in sangria_levels.split(";")] if sangria_levels else []
        if len(sangria_levels_list) == 3:
            amount = float(amount)
            if amount >= sangria_levels_list[2]:
                show_messagebox(pos_id, "$NECESSARY_BLEED", title="$SKIM_ALERT", buttons="$OK")
                return True
            elif state in ["TOTALED", "PAID"] and sangria_levels_list[0] <= amount < sangria_levels_list[2]:
                if amount < sangria_levels_list[1]:
                    if int(alert_level_1 or 0) == 0:
                        set_custom(pos_id, "sangria_level_1_alert", "1")
                    else:
                        return False
                show_messagebox(pos_id, "$NECESSARY_BLEED", title="$SKIM_ALERT", icon="error", buttons="$OK")
    except (BaseException,):
        logger.exception("Error in Skim Level configuration")

    return False


def check_sangria(model, pos_id):
    if is_sangria_enable():
        period = get_business_period(model)
        session = get_operator_session(model)
        drawer_amount = get_drawer_amount(pos_id, period, session)

        logger.debug("--- doSale doSetDrawerStatus ---")
        if do_set_drawer_status(pos_id, drawer_amount):
            raise StopAction()


def is_sangria_enable():
    from posactions import is_sangria_enabled_config, sangria_levels
    return is_sangria_enabled_config and sangria_levels is not None



import json

from helper import json_serialize
from typing import Any

from messagebus import Message, DefaultToken, DataType
from messageprocessorutil import UuidMessageProcessor

from customer_services.interactor.update_customer import UpdateCustomerInteractor
from customer_services.interactor.update_customer.dto import UpdateCustomerRequest

from externals.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)


class UpdateCustomerProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (UpdateCustomerInteractor) -> None
        
        super(UpdateCustomerProcessor, self).__init__()
        
        self.interactor = interactor
    
    def parse_data(self, data):
        # type: (Any) -> UpdateCustomerRequest
        
        request = json.loads(data)
        return UpdateCustomerRequest(request["phone"],
                                     request["document"],
                                     request["name"],
                                     request["zip_code"],
                                     request["city"],
                                     request["street"],
                                     request["neighborhood"],
                                     request["number"],
                                     request["complement"],
                                     request["reference_point"])
    
    def call_business(self, request):
        # type: (UpdateCustomerRequest) -> None

        logger.info("Updating customer for request: {}".format(json_serialize(request)))
        self.interactor.update_customer(request)
    
    def format_response(self, message_bus, message, event, request, response):
        logger.info("Customer updated for request: {}".format(json_serialize(request)))
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.param, response))
    
    def format_exception(self, message_bus, message, event, request, exception):
        exception_name = str(type(exception).__name__)
        logger.info("Error updating customer from request {}: {}".format(json_serialize(request), exception_name))
        reply_message = Message(DefaultToken.TK_SYS_NAK, DataType.param, exception_name)
        message_bus.reply_message(message, reply_message)
    
    def format_parse_exception(self, message_bus, message, event, data, exception):
        logger.info("Error getting parsing request: {}".format(str(exception)))
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from old_helper import BaseRepository
from typing import List, Tuple
from persistence import Connection


class TenderRepository(BaseRepository):
    def __init__(self, mbcontext):
        super(TenderRepository, self).__init__(mbcontext)

    def get_tender_names(self):
        # type: () -> List[Tuple[int, unicode]]
        def inner_func(conn):
            # type: (Connection) -> List[Tuple[int, unicode]]
            return [(int(x.get_entry(0)), x.get_entry(1).decode("utf-8")) for x in conn.select(self._TenderNamesQuery)]

        return self.execute_with_connection(inner_func)

    _TenderNamesQuery = "select TenderId, TenderDescr from TenderType"



from mw_helper import ensure_iterable, convert_to_dict
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates


class OrderSubTypeFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, allowed_subtypes, forbidden_subtypes, logger=None):
        super(OrderSubTypeFilterBox, self).__init__(name, sons, logger)
        self.allowed_subtypes = None
        if allowed_subtypes is not None:
            self.allowed_subtypes = convert_to_dict(ensure_iterable(allowed_subtypes))
        self.forbidden_subtypes = None
        if forbidden_subtypes is not None:
            self.forbidden_subtypes = convert_to_dict(ensure_iterable(forbidden_subtypes))

    def change_order(self, order):
        if self.forbidden_subtypes is not None and order.order_subtype in self.forbidden_subtypes:
            order.prod_state = ProdStates.INVALID
            return order

        if self.allowed_subtypes is not None and order.order_subtype not in self.allowed_subtypes:
            order.prod_state = ProdStates.INVALID
            return order

        return order



from datetime import datetime

from helper import BaseRepository, convert_from_utf_to_localtime
from persistence import Connection  # noqa
from typing import Optional, List  # noqa

from reports_app.cashtransferreport.dto import Transfer


class AccountRepository(BaseRepository):
    def __init__(self, mbcontext):
        super(AccountRepository, self).__init__(mbcontext)

    def get_transfers_by_business_period(self, initial_date, end_date, operator_id, report_pos, report_name):
        # type: (datetime, datetime, Optional[str], Optional[str], Optional[str]) -> List[Transfer]
        return self._get_transfer(initial_date, end_date, operator_id, report_pos, report_name)

    def get_transfers_by_real_date(self, initial_date, end_date, operator_id, report_pos, report_name):
        # type: (datetime, datetime, Optional[str], Optional[str], Optional[str]) -> List[Transfer]
        return self._get_transfer(initial_date, end_date, operator_id, report_pos, report_name)

    def get_transfers_by_session_id(self, session_id, report_name):
        # type: (unicode) -> List[Transfer]
        return self._get_transfer(None, None, None, session_id, report_name)

    def _get_transfer(self, initial_date, end_date, operator_id, report_pos, report_name):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[int], Optional[str], Optional[str]) -> List[Transfer]
        def inner_func(conn):
            # type: (Connection) -> List[Transfer]
            query = self._get_transfer_query(initial_date, end_date, operator_id, report_pos, report_name, report_pos)
            query_result = conn.select(query)
            transfers = []

            for transfer in query_result:
                self.append_transfers_type(transfer, transfers, report_name)

            return transfers

        return self.execute_with_connection(inner_func)

    _TransfersByBusinessPeriod = "select Description, SessionId, Timestamp, Amount, Period, Type " \
                                 "from Transfer where Period >= '{0}' and Period <= '{1}'"

    _TransfersByRealDate = "select Type, Amount from Transfer where strftime('%Y-%m-%d', Timestamp, 'localtime') >= " \
                           "'{0}' and strftime('%Y-%m-%d', Timestamp, 'localtime') <= '{1}'"

    _TransfersBySessionId = "select Type, Amount from Transfer where SessionId = '{0}'"

    def _get_transfer_query(self, initial_date, end_date, operator_id, report_pos, report_name, session_id):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[int], Optional[str], Optional[str]) -> List[Transfer]
        query = ''
        if report_name == "report_by_business_period" or report_name == "cash_in" or report_name == "cash_out" or report_name == "report_by_xml":
            query += self._get_transfer_query_business_period(end_date, initial_date)
        if report_name == "report_by_real_date":
            query += self._get_transfer_query_real_date(end_date, initial_date)
        if report_name == "report_by_session_id":
            query += self._get_transfer_query_session_id(session_id)
        if report_name != "report_by_xml":
            query += self._get_transfer_query_operator_id(operator_id, report_name)
        query += self._get_transfer_query_report_pos(report_pos, report_name)
        query += self._get_transfer_query_cash_out_cash_in(report_name)
        query += " order by Timestamp asc;"
        return query

    def _get_transfer_query_session_id(self, session_id):
        query = self._TransfersBySessionId
        query = query.format(session_id)
        return query

    def _get_transfer_query_real_date(self, end_date, initial_date):
        query = self._TransfersByBusinessPeriod
        query = query.format(initial_date.strftime("%Y%m%d"), end_date.strftime("%Y%m%d"))
        return query

    def _get_transfer_query_business_period(self, end_date, initial_date):
        return self._get_transfer_query_real_date(end_date, initial_date)

    @staticmethod
    def _get_transfer_query_operator_id(operator_id, report_name):
        query = ''
        if operator_id is not None and report_name != 'report_by_session_id':
            query += " and SessionId like '%user={0},%'".format(operator_id)
        return query

    @staticmethod
    def _get_transfer_query_report_pos(report_pos, report_name):
        query = ''
        if report_pos is not None and report_name != 'report_by_session_id':
            query += " and SessionId like '%pos={0},%'".format(report_pos)
        return query

    @staticmethod
    def _get_transfer_query_cash_out_cash_in(report_name):
        query = ''
        if report_name == 'cash_out':
            query += " and Type = 4"
        elif report_name == 'cash_in':
            query += " and (Type = 1 or Type = 3)"
        else:  # cash_in and cash_out
            query += " and (Type = 1 or Type = 3 or Type = 4)"

        return query

    def append_transfers_type(self, transfer, transfers, report_name):
        transfer_type = self._get_transfer_type(transfer)
        transfer_operator = self._get_transfer_operator(transfer, report_name)
        transfer_date = self._get_transfer_date(transfer, report_name)
        if transfer_date is not None:
            transfer_date = convert_from_utf_to_localtime(transfer_date)
        transfer_value = self._get_transfer_value(transfer, report_name)
        transfer_cash_in_cash_out = self._get_transfer_cash_in_cash_out(transfer, report_name)
        transfers.append(Transfer(transfer_type, transfer_operator, transfer_date, transfer_value, transfer_cash_in_cash_out))

    @staticmethod
    def _get_transfer_type(transfer):
        transfer_type = None if transfer.get_entry(5) is None else int(transfer.get_entry(5))
        return AccountRepository._select_transfer_type(transfer, transfer_type)

    @staticmethod
    def _select_transfer_type(transfer, transfer_type):
        cash_in_by_login = 1
        cash_in_by_manager_menu = 3

        if transfer_type == cash_in_by_login:
            return '$LOGIN'
        else:
            transfer_description = str(transfer.get_entry(0))
            if '"type":' in transfer_description:
                description = transfer_description.split('"type":')[1].split(',')[0][2:-1]
                if transfer_type == cash_in_by_manager_menu:
                    description = description[:-1]
                return description
        return ''

    @staticmethod
    def _get_transfer_operator(transfer, report_name):
        if report_name != "report_by_session_id":
            return str(transfer.get_entry(1).split("user=")[1].split(",")[0])
        return None

    @staticmethod
    def _get_transfer_date(transfer, report_name):
        if report_name != "report_by_session_id":
            return datetime.strptime(transfer.get_entry(2), "%Y-%m-%d %H:%M:%S")
        return None

    @staticmethod
    def _get_transfer_value(transfer, report_name):
        if report_name != "report_by_session_id":
            return float(transfer.get_entry(3))
        return float(transfer.get_entry(1))

    @staticmethod
    def _get_transfer_cash_in_cash_out(transfer, report_name):
        if report_name != "report_by_session_id":
            return int(transfer.get_entry(5))
        return int(transfer.get_entry(0))



# -*- coding: utf-8 -*-

import sysactions

from systools import sys_log_exception


from .. import logger

from .. import mb_context


@sysactions.action
def order_ready_to_go(pos_id, order_id, external_id):
    try:
        xml = "|".join([order_id, external_id])
        mb_context.MB_EasyEvtSend("OrderReadyToGo", type="", xml=xml, sourceid=int(pos_id))

    except Exception as e:
        sys_log_exception("Error OrderReadyToGo")
        logger.error(e)
        return False
